import argparse
from collections import deque

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.distributions import Categorical

from ignite.engine import Engine, Events

try:
    import gymnasium as gym
except ImportError:
    raise ModuleNotFoundError("Please install opengym: pip install gymnasium")


eps = np.finfo(np.float32).eps.item()


class Policy(nn.Module):
    def __init__(self):
        super(Policy, self).__init__()
        self.affine1 = nn.Linear(4, 128)
        self.dropout = nn.Dropout(p=0.6)
        self.affine2 = nn.Linear(128, 2)

        self.saved_log_probs = []
        self.rewards = []

    def forward(self, x):
        x = self.affine1(x)
        x = self.dropout(x)
        x = F.relu(x)
        action_scores = self.affine2(x)
        return F.softmax(action_scores, dim=1)


def select_action(policy, observation):
    state = torch.from_numpy(observation).float().unsqueeze(0)
    probs = policy(state)
    m = Categorical(probs)
    action = m.sample()
    policy.saved_log_probs.append(m.log_prob(action))
    return action.item()


def finish_episode(policy, optimizer, gamma):
    R = 0
    policy_loss = []
    returns = deque()
    for r in policy.rewards[::-1]:
        R = r + gamma * R
        returns.appendleft(R)
    returns = torch.tensor(returns)
    returns = (returns - returns.mean()) / (returns.std() + eps)
    for log_prob, R in zip(policy.saved_log_probs, returns):
        policy_loss.append(-log_prob * R)
    optimizer.zero_grad()
    policy_loss = torch.cat(policy_loss).sum()
    policy_loss.backward()
    optimizer.step()
    del policy.rewards[:]
    del policy.saved_log_probs[:]


EPISODE_STARTED = Events.EPOCH_STARTED
EPISODE_COMPLETED = Events.EPOCH_COMPLETED


def main(env, args):
    policy = Policy()
    optimizer = optim.Adam(policy.parameters(), lr=1e-2)
    timesteps = range(10000)

    def run_single_timestep(engine, timestep):
        observation = engine.state.observation
        action = select_action(policy, observation)
        engine.state.observation, reward, done, _, _ = env.step(action)
        if args.render:
            env.render()
        policy.rewards.append(reward)
        engine.state.ep_reward += reward
        if done:
            engine.terminate_epoch()
            engine.state.timestep = timestep

    trainer = Engine(run_single_timestep)
    trainer.state.running_reward = 10

    @trainer.on(EPISODE_STARTED)
    def reset_environment_state():
        torch.manual_seed(args.seed + trainer.state.epoch)
        trainer.state.observation, _ = env.reset(seed=args.seed + trainer.state.epoch)
        trainer.state.ep_reward = 0

    @trainer.on(EPISODE_COMPLETED)
    def update_model():
        trainer.state.running_reward = 0.05 * trainer.state.ep_reward + (1 - 0.05) * trainer.state.running_reward
        finish_episode(policy, optimizer, args.gamma)

    @trainer.on(EPISODE_COMPLETED(every=args.log_interval))
    def log_episode():
        i_episode = trainer.state.epoch
        print(
            f"Episode {i_episode}\tLast reward: {trainer.state.ep_reward:.2f}"
            f"\tAverage length: {trainer.state.running_reward:.2f}"
        )

    @trainer.on(EPISODE_COMPLETED)
    def should_finish_training():
        running_reward = trainer.state.running_reward
        if running_reward > env.spec.reward_threshold:
            print(
                f"Solved! Running reward is now {running_reward} and "
                f"the last episode runs to {trainer.state.timestep} time steps!"
            )
            trainer.should_terminate = True

    trainer.run(timesteps, max_epochs=args.max_episodes)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="PyTorch REINFORCE example")
    parser.add_argument("--gamma", type=float, default=0.99, metavar="G", help="discount factor (default: 0.99)")
    parser.add_argument("--seed", type=int, default=543, metavar="N", help="random seed (default: 543)")
    parser.add_argument("--render", action="store_true", help="render the environment")
    parser.add_argument(
        "--log-interval", type=int, default=10, metavar="N", help="interval between training status logs (default: 10)"
    )
    parser.add_argument(
        "--max-episodes",
        type=int,
        default=1000000,
        metavar="N",
        help="Number of episodes for the training (default: 1000000)",
    )
    args = parser.parse_args()

    env = gym.make("CartPole-v1")

    main(env, args)



import os
import stat
import warnings
from collections import OrderedDict
from collections.abc import Mapping
from pathlib import Path
from unittest.mock import MagicMock

import pytest
import torch
import torch.nn as nn
from packaging.version import Version

import ignite.distributed as idist
from ignite.engine import Engine, Events, State
from ignite.handlers import Checkpoint, DiskSaver, EarlyStopping, global_step_from_engine, ModelCheckpoint
from ignite.handlers.checkpoint import BaseSaveHandler

_PREFIX = "PREFIX"


class DummyModel(nn.Module):
    def __init__(self):
        super(DummyModel, self).__init__()
        self.net = nn.Linear(1, 1)

    def forward(self, x):
        return self.net(x)


model = DummyModel()
optimizer = torch.optim.SGD(model.parameters(), lr=0.1)


class DummyPretrainedModel(nn.Module):
    def __init__(self):
        super(DummyPretrainedModel, self).__init__()
        self.features = nn.Linear(4, 2, bias=False)
        self.fc = nn.Linear(2, 1)

    def forward(self, x):
        x = self.features(x)
        x = self.fc(x)
        return x


def test_checkpoint_wrong_input():
    with pytest.raises(TypeError, match=r"Argument `to_save` should be a dictionary"):
        Checkpoint(12, lambda x: x, "prefix")

    with pytest.raises(TypeError, match=r"Argument `to_save` should be a dictionary"):
        Checkpoint([12], lambda x: x, "prefix")

    with pytest.raises(TypeError, match=r"should have `state_dict`"):
        Checkpoint({"model": {"abc": 12}}, lambda x: x, "prefix")

    to_save = {"model": model}

    with pytest.raises(
        TypeError,
        match=r"Argument `save_handler` should be a string or Path object or callable or inherit from BaseSaveHandler",
    ):
        Checkpoint(to_save, 12, "prefix")

    with pytest.raises(TypeError, match=r"global_step_transform should be a function."):
        Checkpoint(to_save, lambda x: x, score_function=lambda e: 123, score_name="acc", global_step_transform=123)

    with pytest.raises(ValueError, match=r"Cannot have key 'checkpointer' if `include_self` is True"):
        Checkpoint({"checkpointer": model}, lambda x: x, include_self=True)

    class ImmutableMapping(Mapping):
        def __init__(self, d):
            self._dict = d

        def __getitem__(self, key):
            return self._dict[key]

        def __iter__(self):
            return iter(self._dict)

        def __len__(self):
            return len(self._dict)

    with pytest.raises(TypeError, match="If `include_self` is True, then `to_save` must be mutable"):
        Checkpoint(ImmutableMapping(to_save), lambda x: x, include_self=True)

    checkpoint = Checkpoint(to_save, lambda x: x)
    with pytest.raises(AttributeError, match="Checkpoint's `save_handler` should be of type `DiskSaver`"):
        checkpoint.reload_objects(to_save)


def test_save_handler_as_str(dirname):
    to_save = {"model": model}

    checkpointer = Checkpoint(to_save, save_handler=dirname)
    assert isinstance(checkpointer.save_handler, DiskSaver)


def test_checkpoint_score_function_wrong_output():
    to_save = {"model": model}

    checkpointer = Checkpoint(to_save, lambda x: x, score_function=lambda e: {"1": 1}, score_name="acc")
    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)
    with pytest.raises(ValueError, match=r"Output of score_function should be a number"):
        checkpointer(trainer)


@pytest.mark.parametrize(
    "to_save, obj, name",
    [
        ({"model": model}, model.state_dict(), "model"),
        (
            {"model": model, "optimizer": optimizer},
            {"model": model.state_dict(), "optimizer": optimizer.state_dict()},
            "checkpoint",
        ),
    ],
)
def test_checkpoint_default(to_save, obj, name):
    save_handler = MagicMock(spec=BaseSaveHandler)

    checkpointer = Checkpoint(to_save, save_handler=save_handler)
    assert checkpointer.last_checkpoint is None

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)

    checkpointer(trainer)
    assert save_handler.call_count == 1

    metadata = {"basename": name, "score_name": None, "priority": 0}
    save_handler.assert_called_with(obj, f"{name}_0.pt", metadata)

    trainer.state.epoch = 12
    trainer.state.iteration = 1234
    checkpointer(trainer)
    assert save_handler.call_count == 2
    metadata["priority"] = 1234
    save_handler.assert_called_with(obj, f"{name}_1234.pt", metadata)
    assert save_handler.remove.call_count == 1
    save_handler.remove.assert_called_with(f"{name}_0.pt")
    assert checkpointer.last_checkpoint == f"{name}_1234.pt"


@pytest.mark.parametrize(
    "to_save, obj, name",
    [
        ({"model": model}, model.state_dict(), "model"),
        (
            {"model": model, "optimizer": optimizer},
            {"model": model.state_dict(), "optimizer": optimizer.state_dict()},
            "checkpoint",
        ),
    ],
)
def test_checkpoint_include_self_state_dict(to_save, obj, name):
    save_handler = MagicMock(spec=BaseSaveHandler)

    checkpointer = Checkpoint(to_save, save_handler=save_handler, include_self=True)
    assert checkpointer.last_checkpoint is None

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)

    checkpointer(trainer)
    assert save_handler.call_count == 1

    fname = f"{name}_0.pt"
    obj["checkpointer"] = OrderedDict([("_saved", [(0, fname)])])

    metadata = {"basename": name, "score_name": None, "priority": 0}
    save_handler.assert_called_with(obj, fname, metadata)

    # Swap object, state should be maintained
    checkpointer2 = Checkpoint(to_save, save_handler=save_handler, include_self=True)
    checkpointer2.load_state_dict(checkpointer.state_dict())
    assert checkpointer2.last_checkpoint == fname

    trainer.state.epoch = 12
    trainer.state.iteration = 1234
    checkpointer2(trainer)
    assert save_handler.call_count == 2
    metadata["priority"] = 1234

    # This delete only happens if state was restored correctly.
    save_handler.remove.assert_called_with(f"{name}_0.pt")

    fname = f"{name}_1234.pt"
    obj["checkpointer"] = OrderedDict([("_saved", [(1234, fname)])])

    save_handler.assert_called_with(obj, fname, metadata)
    assert save_handler.remove.call_count == 1
    assert checkpointer2.last_checkpoint == fname


def test_checkpoint_with_dp():
    dp_model = nn.DataParallel(model)
    to_save = {"model": dp_model}

    save_handler = MagicMock(spec=BaseSaveHandler)
    checkpointer = Checkpoint(to_save, save_handler=save_handler)

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)

    checkpointer(trainer)
    assert save_handler.call_count == 1
    metadata = {"basename": "model", "score_name": None, "priority": 0}
    save_handler.assert_called_with(model.state_dict(), "model_0.pt", metadata)


@pytest.mark.parametrize("filename_prefix", ["", "dummytask"])
@pytest.mark.parametrize(
    "to_save, obj, name",
    [
        ({"model": model}, model.state_dict(), "model"),
        (
            {"model": model, "optimizer": optimizer},
            {"model": model.state_dict(), "optimizer": optimizer.state_dict()},
            "checkpoint",
        ),
    ],
)
def test_checkpoint_with_global_step_transform(filename_prefix, to_save, obj, name):
    save_handler = MagicMock(spec=BaseSaveHandler)

    checkpointer = Checkpoint(
        to_save,
        save_handler=save_handler,
        filename_prefix=filename_prefix,
        global_step_transform=lambda e, _: e.state.epoch,
    )

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=2, iteration=1)

    checkpointer(trainer)
    assert save_handler.call_count == 1

    if len(filename_prefix) > 0:
        filename_prefix += "_"

    metadata = {"basename": f"{filename_prefix}{name}", "score_name": None, "priority": 2}
    save_handler.assert_called_with(obj, f"{filename_prefix}{name}_2.pt", metadata)

    trainer.state.epoch = 12
    trainer.state.iteration = 1234
    checkpointer(trainer)
    assert save_handler.call_count == 2
    metadata["priority"] = 12
    save_handler.assert_called_with(obj, f"{filename_prefix}{name}_12.pt", metadata)
    assert save_handler.remove.call_count == 1
    save_handler.remove.assert_called_with(f"{filename_prefix}{name}_2.pt")
    assert checkpointer.last_checkpoint == f"{filename_prefix}{name}_12.pt"


@pytest.mark.parametrize(
    "to_save, obj, name",
    [
        ({"model": model}, model.state_dict(), "model"),
        (
            {"model": model, "optimizer": optimizer},
            {"model": model.state_dict(), "optimizer": optimizer.state_dict()},
            "checkpoint",
        ),
    ],
)
def test_checkpoint_with_score_function(to_save, obj, name):
    save_handler = MagicMock(spec=BaseSaveHandler)

    checkpointer = Checkpoint(to_save, save_handler=save_handler, score_function=lambda e: e.state.score)

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=1, iteration=1, score=0.77)

    checkpointer(trainer)
    assert save_handler.call_count == 1

    metadata = {"basename": name, "score_name": None, "priority": 0.77}
    save_handler.assert_called_with(obj, f"{name}_0.7700.pt", metadata)

    trainer.state.epoch = 12
    trainer.state.iteration = 1234
    trainer.state.score = 0.78

    checkpointer(trainer)
    assert save_handler.call_count == 2
    metadata["priority"] = 0.78
    save_handler.assert_called_with(obj, f"{name}_0.7800.pt", metadata)
    assert save_handler.remove.call_count == 1
    save_handler.remove.assert_called_with(f"{name}_0.7700.pt")
    assert checkpointer.last_checkpoint == f"{name}_0.7800.pt"


def test_checkpoint_with_score_name_only():
    to_save = {"model": model}
    obj = model.state_dict()
    name = "model"
    save_handler = MagicMock(spec=BaseSaveHandler)

    trainer = Engine(lambda e, b: None)
    evaluator = Engine(lambda e, b: None)
    trainer.state = State(epoch=11, iteration=1)

    checkpointer = Checkpoint(
        to_save,
        save_handler=save_handler,
        global_step_transform=lambda _1, _2: trainer.state.epoch,
        score_name="val_acc",
    )

    evaluator.state = State(epoch=1, iteration=1000, metrics={"val_acc": 0.77})

    checkpointer(evaluator)
    assert save_handler.call_count == 1

    metadata = {"basename": name, "score_name": "val_acc", "priority": 0.77}
    save_handler.assert_called_with(obj, f"{name}_11_val_acc=0.7700.pt", metadata)

    trainer.state.epoch = 12
    evaluator.state.metrics["val_acc"] = 0.78

    checkpointer(evaluator)
    assert save_handler.call_count == 2
    metadata["priority"] = 0.78
    save_handler.assert_called_with(obj, f"{name}_12_val_acc=0.7800.pt", metadata)
    assert save_handler.remove.call_count == 1
    save_handler.remove.assert_called_with(f"{name}_11_val_acc=0.7700.pt")
    assert checkpointer.last_checkpoint == f"{name}_12_val_acc=0.7800.pt"


@pytest.mark.parametrize(
    "to_save, obj, name",
    [
        ({"model": model}, model.state_dict(), "model"),
        (
            {"model": model, "optimizer": optimizer},
            {"model": model.state_dict(), "optimizer": optimizer.state_dict()},
            "checkpoint",
        ),
    ],
)
def test_checkpoint_with_score_name_and_function(to_save, obj, name):
    save_handler = MagicMock(spec=BaseSaveHandler)

    checkpointer = Checkpoint(
        to_save, save_handler=save_handler, score_name="loss", score_function=lambda e: e.state.score
    )

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=1, iteration=1, score=-0.77)

    checkpointer(trainer)
    assert save_handler.call_count == 1

    metadata = {"basename": name, "score_name": "loss", "priority": -0.77}
    save_handler.assert_called_with(obj, f"{name}_loss=-0.7700.pt", metadata)

    trainer.state.epoch = 12
    trainer.state.iteration = 1234
    trainer.state.score = -0.76

    checkpointer(trainer)
    assert save_handler.call_count == 2
    metadata["priority"] = -0.76
    save_handler.assert_called_with(obj, f"{name}_loss=-0.7600.pt", metadata)
    assert save_handler.remove.call_count == 1
    save_handler.remove.assert_called_with(f"{name}_loss=-0.7700.pt")
    assert checkpointer.last_checkpoint == f"{name}_loss=-0.7600.pt"


def test_checkpoint_with_int_score():
    def _test(to_save, obj, name, score_name=None):
        save_handler = MagicMock(spec=BaseSaveHandler)

        checkpointer = Checkpoint(
            to_save, save_handler=save_handler, score_name=score_name, score_function=lambda e: e.state.epoch
        )

        if score_name is None:
            score_name = ""
        else:
            score_name += "="

        trainer = Engine(lambda e, b: None)
        trainer.state = State(epoch=1, iteration=1)

        checkpointer(trainer)
        assert save_handler.call_count == 1

        metadata = {"basename": name, "score_name": score_name[:-1] if len(score_name) > 0 else None, "priority": 1}
        save_handler.assert_called_with(obj, f"{name}_{score_name}1.pt", metadata)

        trainer.state.epoch = 12
        trainer.state.iteration = 1234

        checkpointer(trainer)
        assert save_handler.call_count == 2
        metadata["priority"] = 12
        save_handler.assert_called_with(obj, f"{name}_{score_name}12.pt", metadata)
        assert save_handler.remove.call_count == 1
        save_handler.remove.assert_called_with(f"{name}_{score_name}1.pt")
        assert checkpointer.last_checkpoint == f"{name}_{score_name}12.pt"

    model = DummyModel()
    to_save = {"model": model}
    _test(to_save, model.state_dict(), "model")
    _test(to_save, model.state_dict(), "model", "epoch")

    model = DummyModel()
    optimizer = torch.optim.SGD(model.parameters(), lr=0.1)
    to_save = {"model": model, "optimizer": optimizer}
    _test(to_save, {"model": model.state_dict(), "optimizer": optimizer.state_dict()}, "checkpoint")
    _test(to_save, {"model": model.state_dict(), "optimizer": optimizer.state_dict()}, "checkpoint", "epoch")


def test_checkpoint_with_score_function_and_trainer_epoch():
    to_save = {"model": model}
    obj = model.state_dict()
    name = "model"
    save_handler = MagicMock(spec=BaseSaveHandler)

    trainer = Engine(lambda e, b: None)
    evaluator = Engine(lambda e, b: None)
    trainer.state = State(epoch=11, iteration=1)

    checkpointer = Checkpoint(
        to_save,
        save_handler=save_handler,
        global_step_transform=lambda _1, _2: trainer.state.epoch,
        score_function=lambda e: e.state.metrics["val_acc"],
    )

    evaluator.state = State(epoch=1, iteration=1000, metrics={"val_acc": 0.77})
    checkpointer(evaluator)
    assert save_handler.call_count == 1

    metadata = {"basename": name, "score_name": None, "priority": 0.77}
    save_handler.assert_called_with(obj, f"{name}_11_0.7700.pt", metadata)

    trainer.state.epoch = 12
    evaluator.state.metrics["val_acc"] = 0.78

    checkpointer(evaluator)
    assert save_handler.call_count == 2
    metadata["priority"] = 0.78
    save_handler.assert_called_with(obj, f"{name}_12_0.7800.pt", metadata)
    assert save_handler.remove.call_count == 1
    save_handler.remove.assert_called_with(f"{name}_11_0.7700.pt")
    assert checkpointer.last_checkpoint == f"{name}_12_0.7800.pt"


def test_checkpoint_with_score_name_and_function_and_trainer_epoch():
    to_save = {"model": model}
    obj = model.state_dict()
    name = "model"
    save_handler = MagicMock(spec=BaseSaveHandler)

    trainer = Engine(lambda e, b: None)
    evaluator = Engine(lambda e, b: None)
    trainer.state = State(epoch=11, iteration=1)

    checkpointer = Checkpoint(
        to_save,
        save_handler=save_handler,
        global_step_transform=lambda _1, _2: trainer.state.epoch,
        score_name="val_acc",
        score_function=lambda e: e.state.metrics["val_acc"],
    )

    evaluator.state = State(epoch=1, iteration=1000, metrics={"val_acc": 0.77})

    checkpointer(evaluator)
    assert save_handler.call_count == 1

    metadata = {"basename": name, "score_name": "val_acc", "priority": 0.77}
    save_handler.assert_called_with(obj, f"{name}_11_val_acc=0.7700.pt", metadata)

    trainer.state.epoch = 12
    evaluator.state.metrics["val_acc"] = 0.78

    checkpointer(evaluator)
    assert save_handler.call_count == 2
    metadata["priority"] = 0.78
    save_handler.assert_called_with(obj, f"{name}_12_val_acc=0.7800.pt", metadata)
    assert save_handler.remove.call_count == 1
    save_handler.remove.assert_called_with(f"{name}_11_val_acc=0.7700.pt")
    assert checkpointer.last_checkpoint == f"{name}_12_val_acc=0.7800.pt"


def test_checkpoint_last_checkpoint():
    save_handler = MagicMock(spec=BaseSaveHandler)
    to_save = {"model": DummyModel()}

    checkpointer = Checkpoint(to_save, save_handler=save_handler, n_saved=None)

    trainer = Engine(lambda e, b: None)

    for i in range(10):
        trainer.state = State(epoch=1, iteration=i)
        checkpointer(trainer)

    assert save_handler.call_count == 10
    assert checkpointer.last_checkpoint == "model_9.pt"


def test_checkpoint_last_checkpoint_on_score():
    save_handler = MagicMock(spec=BaseSaveHandler)
    to_save = {"model": DummyModel()}

    checkpointer = Checkpoint(
        to_save,
        save_handler=save_handler,
        n_saved=None,
        score_name="val_acc",
        score_function=lambda e: e.state.metrics["val_acc"],
    )

    trainer = Engine(lambda e, b: None)

    val_acc = 0.0
    for i in range(10):
        val_acc = i * 0.1
        trainer.state = State(epoch=1, iteration=i, metrics={"val_acc": val_acc})
        checkpointer(trainer)

    assert save_handler.call_count == 10
    assert checkpointer.last_checkpoint == "model_val_acc=0.9000.pt"


def test_checkpoint_save_handler_callable():
    def save_handler(c, f):
        assert f == "model_12.pt"

    to_save = {"model": DummyModel()}

    checkpointer = Checkpoint(to_save, save_handler=save_handler)

    trainer = Engine(lambda e, b: None)

    trainer.state = State(epoch=1, iteration=12)
    checkpointer(trainer)


def test_model_checkpoint_args_validation(dirname):
    existing = dirname / "existing_dir"
    nonempty = dirname / "nonempty"

    existing.mkdir(parents=True)
    nonempty.mkdir(parents=True)

    with open(nonempty / f"{_PREFIX}_name_0.pt", "w"):
        pass

    with pytest.raises(ValueError, match=r"with extension '.pt' are already present "):
        ModelCheckpoint(nonempty, _PREFIX)

    with pytest.raises(ValueError, match=r"Directory path '\S+' is not found"):
        ModelCheckpoint(dirname / "non_existing_dir", _PREFIX, create_dir=False)

    with pytest.raises(TypeError, match=r"global_step_transform should be a function"):
        ModelCheckpoint(existing, _PREFIX, create_dir=False, global_step_transform=1234)

    h = ModelCheckpoint(dirname, _PREFIX, create_dir=False)
    assert h.last_checkpoint is None
    with pytest.raises(RuntimeError, match=r"No objects to checkpoint found."):
        h(None, [])


def test_model_checkpoint_simple_recovery(dirname):
    h = ModelCheckpoint(dirname, _PREFIX, create_dir=False)
    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=1)

    model = DummyModel()
    to_save = {"model": model}
    h(engine, to_save)

    fname = h.last_checkpoint
    assert isinstance(fname, Path)
    assert str(dirname / _PREFIX) in str(fname)
    assert fname.exists()
    loaded_objects = torch.load(fname)
    assert loaded_objects == model.state_dict()
    to_load = {"model": DummyModel()}
    h.reload_objects(to_load=to_load, global_step=1)
    assert to_load["model"].state_dict() == model.state_dict()


@pytest.mark.parametrize("ext, require_empty", [(".txt", True), (".pt", False)])
def test_model_checkpoint_simple_recovery_from_existing_non_empty(ext, require_empty, dirname):
    previous_fname = dirname / f"{_PREFIX}_obj_{1}{ext}"
    with open(previous_fname, "w") as f:
        f.write("test")

    h = ModelCheckpoint(dirname, _PREFIX, create_dir=True, require_empty=require_empty)
    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=1)

    to_save = {"model": model}
    h(engine, to_save)

    fname = h.last_checkpoint
    ext = ".pt"
    assert isinstance(fname, Path)
    assert dirname / f"{_PREFIX}_model_{1}{ext}" == fname
    assert fname.exists()
    assert previous_fname.exists()
    loaded_objects = torch.load(fname)
    assert loaded_objects == model.state_dict()
    to_load = {"model": DummyModel()}
    h.reload_objects(to_load=to_load, global_step=1)
    assert to_load["model"].state_dict() == model.state_dict()
    fname.unlink()


def test_model_checkpoint_invalid_save_handler(dirname):
    h = ModelCheckpoint(dirname, _PREFIX)
    to_save = {"model": DummyModel()}
    # Redefine save_handler
    h.save_handler = lambda x, y: None
    h(Engine(lambda x, y: None), to_save)

    with pytest.raises(
        RuntimeError, match=rf"Internal error, save_handler should be DiskSaver, but has {type(h.save_handler)}."
    ):
        h.last_checkpoint


def test_disk_saver_atomic(dirname):
    model = DummyModel()
    to_save_serializable = {"model": model}
    to_save_non_serializable = {"model": lambda x: x}

    def _test_existence(atomic, _to_save, expected):
        saver = DiskSaver(dirname, atomic=atomic, create_dir=False, require_empty=False)
        fname = "test.pt"
        try:
            with warnings.catch_warnings():
                # Ignore torch/serialization.py:292: UserWarning: Couldn't retrieve source code for container of type
                # DummyModel. It won't be checked for correctness upon loading.
                warnings.simplefilter("ignore", category=UserWarning)
                saver(_to_save, fname)
        except Exception:
            pass
        fp = saver.dirname / fname
        assert fp.exists() == expected

        if expected:
            # related to https://github.com/pytorch/ignite/issues/1876
            mode = stat.filemode(fp.stat().st_mode)
            assert [mode[1], mode[4], mode[7]] == ["r", "r", "r"], mode

        if expected:
            saver.remove(fname)

    _test_existence(atomic=False, _to_save=to_save_serializable, expected=True)
    _test_existence(atomic=False, _to_save=to_save_non_serializable, expected=True)

    _test_existence(atomic=True, _to_save=to_save_serializable, expected=True)
    _test_existence(atomic=True, _to_save=to_save_non_serializable, expected=False)


@pytest.mark.skipif(
    Version(torch.__version__) < Version("1.4.0"), reason="Zipfile serialization was introduced in 1.4.0"
)
def test_disk_saver_zipfile_serialization_keyword(dirname):
    model = DummyModel()
    to_save = {"model": model}

    saver = DiskSaver(dirname, create_dir=False, _use_new_zipfile_serialization=False)
    fname = "test.pt"
    saver(to_save, fname)
    fp = saver.dirname / fname
    assert fp.exists()
    saver.remove(fname)


def test_disk_saver_unknown_keyword(dirname):
    model = DummyModel()
    to_save = {"model": model}

    saver = DiskSaver(dirname, create_dir=False, unknown_keyword="")
    fname = "test.pt"
    with pytest.raises(TypeError, match=r"got an unexpected keyword argument 'unknown_keyword'"):
        saver(to_save, fname)


def test_last_k(dirname):
    h = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=2)
    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=0)

    model = DummyModel()
    to_save = {"model": model}
    h(engine, to_save)

    for i in range(1, 9):
        engine.state.iteration = i
        h(engine, to_save)

    expected = [f"{_PREFIX}_model_{i}.pt" for i in [7, 8]]

    assert sorted(os.listdir(dirname)) == expected, f"{sorted(os.listdir(dirname))} vs {expected}"


def test_disabled_n_saved(dirname):
    h = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=None)
    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=0)

    model = DummyModel()
    to_save = {"model": model}

    num_iters = 100
    for i in range(num_iters):
        engine.state.iteration = i
        h(engine, to_save)

    saved_files = sorted(os.listdir(dirname))
    assert len(saved_files) == num_iters, f"{saved_files}"

    expected = sorted([f"{_PREFIX}_model_{i}.pt" for i in range(num_iters)])
    assert saved_files == expected, f"{saved_files} vs {expected}"


def test_best_k(dirname):
    scores = iter([1.2, -2.0, 3.1, -4.0])

    def score_function(_):
        return next(scores)

    h = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=2, score_function=score_function)

    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=0)

    model = DummyModel()
    to_save = {"model": model}
    for _ in range(4):
        h(engine, to_save)

    expected = [f"{_PREFIX}_model_{i:.4f}.pt" for i in [1.2, 3.1]]

    assert sorted(os.listdir(dirname)) == expected


def test_best_k_with_suffix(dirname):
    scores = [0.3456789, 0.1234, 0.4567, 0.134567]
    scores_iter = iter(scores)

    def score_function(engine):
        return next(scores_iter)

    h = ModelCheckpoint(
        dirname, _PREFIX, create_dir=False, n_saved=2, score_function=score_function, score_name="val_loss"
    )

    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=0)

    model = DummyModel()
    to_save = {"model": model}
    for _ in range(4):
        engine.state.epoch += 1
        h(engine, to_save)

    expected = [f"{_PREFIX}_model_val_loss={scores[e - 1]:.4}.pt" for e in [1, 3]]

    assert sorted(os.listdir(dirname)) == expected


def test_removes_each_score_at_most_once(dirname):
    scores = [0, 1, 1, 2, 3]
    scores_iter = iter(scores)

    def score_function(_):
        return next(scores_iter)

    h = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=2, score_function=score_function)

    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=0)

    model = DummyModel()
    to_save = {"model": model}
    for _ in range(len(scores)):
        h(engine, to_save)

    # If a score was removed multiple times, the code above would have raise a
    # FileNotFoundError. So this just tests the absence of such a failure
    # without futher assertions.


def test_with_engine(dirname):
    def update_fn(_1, _2):
        pass

    name = "model"
    engine = Engine(update_fn)
    handler = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=2)

    model = DummyModel()
    to_save = {"model": model}
    engine.add_event_handler(Events.EPOCH_COMPLETED, handler, to_save)
    engine.run([0, 1], max_epochs=4)

    expected = sorted([f"{_PREFIX}_{name}_{i}.pt" for i in [3 * 2, 4 * 2]])

    assert sorted(os.listdir(dirname)) == expected


def test_with_state_dict(dirname):
    def update_fn(_1, _2):
        pass

    engine = Engine(update_fn)
    handler = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=1)

    model = DummyModel()
    to_save = {"model": model}
    engine.add_event_handler(Events.EPOCH_COMPLETED, handler, to_save)
    engine.run([0, 1, 2], max_epochs=4)

    saved_model = dirname / os.listdir(dirname)[0]
    load_model = torch.load(saved_model)

    assert not isinstance(load_model, DummyModel)
    assert isinstance(load_model, dict)

    model_state_dict = model.state_dict()
    loaded_model_state_dict = load_model
    for key in model_state_dict.keys():
        assert key in loaded_model_state_dict

        model_value = model_state_dict[key]
        loaded_model_value = loaded_model_state_dict[key]

        assert model_value.numpy() == loaded_model_value.numpy()


def test_valid_state_dict_save(dirname):
    model = DummyModel()
    h = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=1)

    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=0)

    to_save = {"name": 42}
    with pytest.raises(TypeError, match=r"should have `state_dict` method"):
        h(engine, to_save)
    to_save = {"name": model}
    try:
        h(engine, to_save)
    except ValueError:
        pytest.fail("Unexpected ValueError")


def _test_save_model_optimizer_lr_scheduler_with_state_dict(device, dirname, just_on_zero_rank=False):
    torch.manual_seed(23)

    model = DummyModel().to(device)

    optim = torch.optim.SGD(model.parameters(), lr=0.1)
    lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optim, gamma=0.5)

    def update_fn(engine, batch):
        x = torch.rand((4, 1)).to(device)
        optim.zero_grad()
        y = model(x)
        # Below code raises: RuntimeError: torch_xla/csrc/tensor_impl.cpp:144 : XLA tensors do not have storage
        # Probably related to https://github.com/pytorch/xla/issues/2576
        # loss = y.pow(2.0).sum()
        loss = y.sum()
        loss.backward()
        if idist.has_xla_support:
            import torch_xla.core.xla_model as xm

            xm.optimizer_step(optim, barrier=True)
        else:
            optim.step()
        lr_scheduler.step()

    engine = Engine(update_fn)

    if (not just_on_zero_rank) or (just_on_zero_rank and idist.get_rank() == 0):
        handler = ModelCheckpoint(dirname, _PREFIX, create_dir=True, n_saved=1)

        engine.add_event_handler(
            Events.EPOCH_COMPLETED, handler, {"model": model, "optimizer": optim, "lr_scheduler": lr_scheduler}
        )

    engine.run([0, 1, 2], max_epochs=4)

    idist.barrier()

    saved_objects = sorted(os.listdir(dirname))
    # saved object is ['PREFIX_checkpoint_3.pt', ]
    saved_checkpoint = dirname / saved_objects[0]

    if idist.has_xla_support:
        device = "cpu"

    loaded_obj = torch.load(saved_checkpoint, map_location=device)
    for f in ["model", "optimizer", "lr_scheduler"]:
        assert f in loaded_obj
    loaded_model_state_dict = loaded_obj["model"]
    loaded_optimizer_state_dict = loaded_obj["optimizer"]
    loaded_lr_scheduler_state_dict = loaded_obj["lr_scheduler"]

    assert isinstance(loaded_model_state_dict, dict)
    assert isinstance(loaded_optimizer_state_dict, dict)
    assert isinstance(loaded_lr_scheduler_state_dict, dict)

    # Specifically move device to CPU first
    model_state_dict = model.cpu().state_dict()
    for key in model_state_dict.keys():
        assert key in loaded_model_state_dict
        model_value = model_state_dict[key]
        loaded_model_value = loaded_model_state_dict[key]
        assert model_value.cpu().numpy() == loaded_model_value.cpu().numpy()

    optim_state_dict = optim.state_dict()
    for key in optim_state_dict.keys():
        assert key in loaded_optimizer_state_dict
        optim_value = optim_state_dict[key]
        loaded_optim_value = loaded_optimizer_state_dict[key]
        if idist.get_rank() == 0:
            assert optim_value == loaded_optim_value

    lr_scheduler_state_dict = lr_scheduler.state_dict()
    for key in lr_scheduler_state_dict.keys():
        assert key in loaded_lr_scheduler_state_dict
        lr_scheduler_value = lr_scheduler_state_dict[key]
        loaded_lr_scheduler_value = loaded_lr_scheduler_state_dict[key]
        assert lr_scheduler_value == loaded_lr_scheduler_value


def test_save_model_optimizer_lr_scheduler_with_state_dict(dirname):
    _test_save_model_optimizer_lr_scheduler_with_state_dict("cpu", dirname)


def _test_save_model_optimizer_lr_scheduler_with_validation(device, dirname, just_on_zero_rank=False):
    torch.manual_seed(23)

    def _build_objects(acc_list):
        model = DummyModel().to(device)
        optim = torch.optim.SGD(model.parameters(), lr=0.1)
        lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optim, gamma=0.5)

        def update_fn(engine, batch):
            x = torch.rand((4, 1)).to(device)
            optim.zero_grad()
            y = model(x)
            loss = y.pow(2.0).sum()
            loss.backward()
            if idist.has_xla_support:
                import torch_xla.core.xla_model as xm

                xm.optimizer_step(optim, barrier=True)
            else:
                optim.step()
            lr_scheduler.step()

        trainer = Engine(update_fn)

        evaluator = Engine(lambda e, b: None)
        acc_iter = iter(acc_list)

        @evaluator.on(Events.EPOCH_COMPLETED)
        def setup_result():
            evaluator.state.metrics["accuracy"] = next(acc_iter)

        @trainer.on(Events.EPOCH_COMPLETED)
        def run_eval():
            evaluator.run([0, 1, 2])

        def score_function(engine):
            return engine.state.metrics["accuracy"]

        save_handler = DiskSaver(dirname, create_dir=True, require_empty=False)
        early_stop = EarlyStopping(score_function=score_function, patience=2, trainer=trainer)
        evaluator.add_event_handler(Events.COMPLETED, early_stop)

        checkpointer = Checkpoint(
            {
                "trainer": trainer,
                "model": model,
                "optim": optim,
                "lr_scheduler": lr_scheduler,
                "early_stop": early_stop,
            },
            save_handler,
            include_self=True,
            global_step_transform=global_step_from_engine(trainer),
        )
        evaluator.add_event_handler(Events.COMPLETED, checkpointer)

        return trainer, evaluator, model, optim, lr_scheduler, early_stop, checkpointer

    trainer, evaluator, model, optim, scheduler, early, checkpointer = _build_objects([0.2, 0.3, 0.2])
    trainer.run([0, 1, 2], max_epochs=3)

    saved_objects = sorted(os.listdir(dirname))
    saved_checkpoint = dirname / saved_objects[0]

    loaded_obj = torch.load(saved_checkpoint, map_location=device)
    for f in ["trainer", "model", "optim", "lr_scheduler", "early_stop", "checkpointer"]:
        assert f in loaded_obj

    trainer2, evaluator2, model2, optim2, scheduler2, early2, checkpointer2 = _build_objects([0.1, 0.1, 0.1])
    Checkpoint.load_objects(
        {
            "trainer": trainer2,
            "model": model2,
            "optim": optim2,
            "lr_scheduler": scheduler2,
            "early_stop": early2,
            "checkpointer": checkpointer2,
        },
        loaded_obj,
    )
    assert checkpointer2.last_checkpoint == checkpointer.last_checkpoint

    model_state_dict = model.cpu().state_dict()
    loaded_model_state_dict = model2.cpu().state_dict()
    for key in model_state_dict.keys():
        assert key in loaded_model_state_dict
        model_value = model_state_dict[key]
        loaded_model_value = loaded_model_state_dict[key]
        assert model_value.cpu().numpy() == loaded_model_value.cpu().numpy()

    optim_state_dict = optim.state_dict()
    loaded_optimizer_state_dict = optim2.state_dict()
    # "params" contains tensor IDs, which are different
    del optim_state_dict["param_groups"][0]["params"]
    del loaded_optimizer_state_dict["param_groups"][0]["params"]
    for key in optim_state_dict.keys():
        assert key in loaded_optimizer_state_dict
        optim_value = optim_state_dict[key]
        loaded_optim_value = loaded_optimizer_state_dict[key]
        if idist.get_rank() == 0:
            assert optim_value == loaded_optim_value

    def _check_state_dict(original, loaded):
        original_state_dict = original.state_dict()
        loaded_state_dict = loaded.state_dict()
        for key in original_state_dict.keys():
            assert key in loaded_state_dict
            original_value = original_state_dict[key]
            loaded_value = loaded_state_dict[key]
            assert original_value == loaded_value

    _check_state_dict(trainer, trainer2)
    _check_state_dict(scheduler, scheduler2)
    _check_state_dict(early, early2)
    _check_state_dict(checkpointer, checkpointer2)

    trainer2.run([0, 1, 2], max_epochs=6)

    # early stopping should have triggered
    assert trainer2.state.epoch == 4

    # If Checkpoint's state was restored correctly, it should continue to respect n_saved
    # and delete old checkpoints, and have the correct last_checkpoint.
    assert os.listdir(dirname) == ["checkpoint_4.pt"]
    assert checkpointer2.last_checkpoint == dirname / "checkpoint_4.pt"


def test_save_model_optimizer_lr_scheduler_with_validation(dirname):
    _test_save_model_optimizer_lr_scheduler_with_validation("cpu", dirname)


def test_checkpoint_load_objects():
    with pytest.raises(TypeError, match=r"Argument checkpoint should be a string or a dictionary"):
        Checkpoint.load_objects({}, [])

    with pytest.raises(TypeError, match=r"should have `load_state_dict` method"):
        Checkpoint.load_objects({"a": None}, {"a": None})

    with pytest.raises(TypeError, match=r"should have `load_state_dict` method"):
        Checkpoint.load_objects({"a": {"b": None}}, {"a": {"b": None}})

    model = DummyModel()
    to_load = {"model": model, "another_model": model}

    with pytest.raises(ValueError, match=r"Key 'model' from x is not found in y"):
        Checkpoint.load_objects(to_load, {})

    model = DummyModel()
    to_load = {"model": model}
    model2 = DummyModel()

    chkpt = {"model": model2.state_dict()}
    Checkpoint.load_objects(to_load, chkpt)
    assert model.state_dict() == model2.state_dict()

    chkpt = {"models": [{"model1": {"abc": model.state_dict()}}, model.state_dict()]}
    to_load = {"models": [{"model1": {"abc": model}}, model]}
    Checkpoint.load_objects(to_load, chkpt)
    assert model.state_dict() == model2.state_dict()


def test_checkpoint_load_objects_from_saved_file(dirname):
    def _get_single_obj_to_save():
        model = DummyModel()
        to_save = {"model": model}
        return to_save

    def _get_multiple_objs_to_save():
        model = DummyModel()
        optim = torch.optim.SGD(model.parameters(), lr=0.001)
        lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optim, gamma=0.5)
        to_save = {
            "model": model,
            "optimizer": optim,
            "lr_scheduler": lr_scheduler,
        }
        return to_save

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)

    # case: load from filepath
    handler = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=1)
    to_save = _get_multiple_objs_to_save()
    handler(trainer, to_save)
    fname = handler.last_checkpoint
    assert isinstance(fname, Path)
    assert str(dirname / _PREFIX) in str(fname)
    assert fname.exists()
    Checkpoint.load_objects(to_save, str(fname))
    Checkpoint.load_objects(to_save, fname)
    fname.unlink()

    # case: multiple objects
    handler = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=1)
    to_save = _get_multiple_objs_to_save()
    handler(trainer, to_save)
    fname = handler.last_checkpoint
    assert isinstance(fname, Path)
    assert str(dirname / _PREFIX) in str(fname)
    assert fname.exists()
    loaded_objects = torch.load(fname)
    Checkpoint.load_objects(to_save, loaded_objects)
    fname.unlink()

    # case: saved multiple objects, loaded single object
    handler = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=1)
    to_save = _get_multiple_objs_to_save()
    handler(trainer, to_save)
    fname = handler.last_checkpoint
    assert isinstance(fname, Path)
    assert str(dirname / _PREFIX) in str(fname)
    assert fname.exists()
    loaded_objects = torch.load(fname)
    to_load = {"model": to_save["model"]}
    Checkpoint.load_objects(to_load, loaded_objects)
    fname.unlink()

    # case: single object
    handler = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=1)
    to_save = _get_single_obj_to_save()
    handler(trainer, to_save)
    fname = handler.last_checkpoint
    assert isinstance(fname, Path)
    assert str(dirname / _PREFIX) in str(fname)
    assert fname.exists()
    loaded_objects = torch.load(fname)
    Checkpoint.load_objects(to_save, loaded_objects)
    fname.unlink()


def test_load_checkpoint_with_different_num_classes(dirname):
    model = DummyPretrainedModel()
    to_save_single_object = {"model": model}

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)

    handler = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=1)
    handler(trainer, to_save_single_object)

    fname = handler.last_checkpoint
    loaded_checkpoint = torch.load(fname)

    to_load_single_object = {"pretrained_features": model.features}

    with pytest.raises(RuntimeError):
        Checkpoint.load_objects(to_load_single_object, loaded_checkpoint)

    Checkpoint.load_objects(to_load_single_object, loaded_checkpoint, strict=False)

    loaded_weights = to_load_single_object["pretrained_features"].state_dict()["weight"]

    assert torch.all(model.state_dict()["features.weight"].eq(loaded_weights))


def test_disksaver_wrong_input(dirname):
    with pytest.raises(ValueError, match=r"Directory path '\S+' is not found"):
        DiskSaver("/tmp/non-existing-folder", create_dir=False)

    def _test(ext):
        previous_fname = dirname / f"{_PREFIX}_obj_{1}{ext}"
        with open(previous_fname, "w") as f:
            f.write("test")

        with pytest.raises(ValueError, match=r"with extension '.pt' are already present"):
            DiskSaver(dirname, require_empty=True)

    _test(".pt")


def _test_checkpoint_with_ddp(device):
    torch.manual_seed(0)

    model = DummyModel().to(device)
    device_ids = None if "cpu" in device.type else [device]
    ddp_model = nn.parallel.DistributedDataParallel(model, device_ids=device_ids)
    to_save = {"model": ddp_model}

    save_handler = MagicMock(spec=BaseSaveHandler)
    checkpointer = Checkpoint(to_save, save_handler=save_handler)

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)

    checkpointer(trainer)
    assert save_handler.call_count == 1
    metadata = {"basename": "model", "score_name": None, "priority": 0}
    save_handler.assert_called_with(model.state_dict(), "model_0.pt", metadata)


def _test_checkpoint_load_objects_ddp(device):
    model = DummyModel().to(device)
    device_ids = None if "cpu" in device.type else [device]
    ddp_model = nn.parallel.DistributedDataParallel(model, device_ids=device_ids)
    opt = torch.optim.SGD(ddp_model.parameters(), lr=0.01)

    # single object:
    to_load = {"model": ddp_model}
    checkpoint = ddp_model.module.state_dict()
    Checkpoint.load_objects(to_load, checkpoint)

    # multiple objects:
    to_load = {"model": ddp_model, "opt": opt}
    checkpoint = {"model": ddp_model.module.state_dict(), "opt": opt.state_dict()}
    Checkpoint.load_objects(to_load, checkpoint)


def _test_checkpoint_with_ZeRO(device, dirname, local_rank):
    from torch.distributed.optim import ZeroRedundancyOptimizer

    model = DummyModel().to(device)
    opt = ZeroRedundancyOptimizer(model.parameters(), torch.optim.SGD, lr=0.01)
    mocked_opt = MagicMock(ZeroRedundancyOptimizer, wraps=opt)

    # A `step` should be called to optimizer state get populated.
    out = model(torch.tensor([1.0], device=device))
    out.backward()
    mocked_opt.step()

    to_save = {"model": model, "optim": mocked_opt}
    checkpointer = Checkpoint(to_save, dirname, save_on_rank=1)

    engine = Engine(lambda e, b: None)
    checkpointer(engine)

    mocked_opt.consolidate_state_dict.assert_called_once_with(to=1)

    if local_rank == 1:
        loaded_state_dict = torch.load(dirname / "checkpoint_0.pt", map_location=device)["optim"]
        state_dict = opt.state_dict()
        assert loaded_state_dict == state_dict


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo, dirname, get_rank_zero_dirname, local_rank):
    device = idist.device()
    rank_zero_dirname = get_rank_zero_dirname()
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device, rank_zero_dirname / "1")
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device, rank_zero_dirname / "2", just_on_zero_rank=True)
    _test_checkpoint_with_ddp(device)
    _test_checkpoint_load_objects_ddp(device)

    from ignite.handlers.checkpoint import HAVE_ZERO

    if HAVE_ZERO:
        _test_checkpoint_with_ZeRO(device, dirname, local_rank)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl, get_rank_zero_dirname):
    device = idist.device()
    dirname = get_rank_zero_dirname()
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device, dirname / "1")
    _test_save_model_optimizer_lr_scheduler_with_state_dict("cpu", dirname / "2", just_on_zero_rank=True)
    _test_checkpoint_with_ddp(device=device)
    _test_checkpoint_load_objects_ddp(device=device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor, get_rank_zero_dirname):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()
    dirname = get_rank_zero_dirname()

    gloo_hvd_executor(
        _test_save_model_optimizer_lr_scheduler_with_state_dict,
        (device, dirname / "1"),
        np=nproc,
        do_init=True,
    )
    gloo_hvd_executor(
        _test_save_model_optimizer_lr_scheduler_with_state_dict,
        ("cpu", dirname / "2", True),
        np=nproc,
        do_init=True,
    )


def _test_tpu_saves_to_cpu(device, dirname):
    torch.manual_seed(0)

    h = ModelCheckpoint(dirname, _PREFIX)
    engine = Engine(lambda e, b: None)
    engine.state = State(epoch=0, iteration=1)

    model = DummyModel().to(device)
    to_save = {"model": model}

    h(engine, to_save)

    idist.barrier()

    fname = h.last_checkpoint
    assert isinstance(fname, Path)
    assert str(dirname / _PREFIX) in str(fname)
    assert fname.exists()
    loaded_objects = torch.load(fname)
    assert loaded_objects == model.cpu().state_dict()


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Not on TPU device")
def test_distrib_single_device_xla(dirname):
    assert "xla" in idist.device().type
    _test_tpu_saves_to_cpu(idist.device(), dirname / "1")
    _test_save_model_optimizer_lr_scheduler_with_state_dict(idist.device(), dirname / "2")


def _test_tpu_saves_to_cpu_nprocs(index, dirname):
    device = idist.device()
    _test_tpu_saves_to_cpu(device, dirname / "1")
    _test_save_model_optimizer_lr_scheduler_with_state_dict(device, dirname / "2")

    import time

    # hack to have all proc properly sync:
    time.sleep(1)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Not on TPU device")
def test_distrib_xla_nprocs(xmp_executor, dirname):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_tpu_saves_to_cpu_nprocs, args=(dirname,), nprocs=n)


def _test_checkpoint_filename_pattern_helper(
    to_save,
    filename_prefix="",
    score_function=None,
    score_name=None,
    global_step_transform=None,
    filename_pattern=None,
    dirname=None,
):
    save_handler = MagicMock(spec=BaseSaveHandler)

    checkpointer = Checkpoint(
        to_save,
        save_handler=save_handler,
        filename_prefix=filename_prefix,
        score_function=score_function,
        score_name=score_name,
        global_step_transform=global_step_transform,
        filename_pattern=filename_pattern,
    )

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=12, iteration=203, score=0.9999)

    checkpointer(trainer)
    return checkpointer.last_checkpoint


def _test_model_checkpoint_filename_pattern_helper(
    to_save,
    filename_prefix="",
    score_function=None,
    score_name=None,
    global_step_transform=None,
    filename_pattern=None,
    dirname=None,
):
    checkpointer = ModelCheckpoint(
        dirname=dirname,
        filename_prefix=filename_prefix,
        score_function=score_function,
        score_name=score_name,
        global_step_transform=global_step_transform,
        filename_pattern=filename_pattern,
        require_empty=False,
    )

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=12, iteration=203, score=0.9999)

    checkpointer(trainer, to_save)
    return Path(checkpointer.last_checkpoint).name


@pytest.mark.parametrize("test_class", ["checkpoint", "model_checkpoint"])
def test_checkpoint_filename_pattern(test_class, dirname):
    if test_class == "checkpoint":
        _test = _test_checkpoint_filename_pattern_helper
    elif test_class == "model_checkpoint":
        _test = _test_model_checkpoint_filename_pattern_helper

    model = DummyModel()
    to_save = {"model": model}

    assert _test(to_save, dirname=dirname) == "model_203.pt"
    assert _test(to_save, "best", dirname=dirname) == "best_model_203.pt"
    assert _test(to_save, score_function=lambda e: e.state.score, dirname=dirname) == "model_0.9999.pt"

    res = _test(
        to_save,
        score_function=lambda e: e.state.score,
        global_step_transform=lambda e, _: e.state.epoch,
        dirname=dirname,
    )
    assert res == "model_12_0.9999.pt"

    assert (
        _test(to_save, score_function=lambda e: e.state.score, score_name="acc", dirname=dirname)
        == "model_acc=0.9999.pt"
    )

    res = _test(
        to_save,
        score_function=lambda e: e.state.score,
        score_name="acc",
        global_step_transform=lambda e, _: e.state.epoch,
        dirname=dirname,
    )
    assert res == "model_12_acc=0.9999.pt"

    assert _test(to_save, "best", score_function=lambda e: e.state.score, dirname=dirname) == "best_model_0.9999.pt"

    res = _test(
        to_save,
        "best",
        score_function=lambda e: e.state.score,
        global_step_transform=lambda e, _: e.state.epoch,
        dirname=dirname,
    )
    assert res == "best_model_12_0.9999.pt"

    res = _test(to_save, "best", score_function=lambda e: e.state.score, score_name="acc", dirname=dirname)
    assert res == "best_model_acc=0.9999.pt"

    res = _test(
        to_save,
        "best",
        score_function=lambda e: e.state.score,
        score_name="acc",
        global_step_transform=lambda e, _: e.state.epoch,
        dirname=dirname,
    )
    assert res == "best_model_12_acc=0.9999.pt"

    pattern = "{name}.{ext}"
    assert _test(to_save, filename_pattern=pattern, dirname=dirname) == "model.pt"

    pattern = "chk-{name}--{global_step}.{ext}"
    assert _test(to_save, to_save, filename_pattern=pattern, dirname=dirname) == "chk-model--203.pt"
    pattern = "chk-{filename_prefix}--{name}--{global_step}.{ext}"
    assert _test(to_save, "best", filename_pattern=pattern, dirname=dirname) == "chk-best--model--203.pt"
    pattern = "chk-{name}--{score}.{ext}"
    assert (
        _test(to_save, score_function=lambda e: e.state.score, filename_pattern=pattern, dirname=dirname)
        == "chk-model--0.9999.pt"
    )
    pattern = "{global_step}-{name}-{score}.chk.{ext}"
    res = _test(
        to_save,
        score_function=lambda e: e.state.score,
        global_step_transform=lambda e, _: e.state.epoch,
        filename_pattern=pattern,
        dirname=dirname,
    )
    assert res == "12-model-0.9999.chk.pt"

    pattern = "chk-{name}--{score_name}--{score}.{ext}"
    res = _test(
        to_save, score_function=lambda e: e.state.score, score_name="acc", filename_pattern=pattern, dirname=dirname
    )
    assert res == "chk-model--acc--0.9999.pt"

    pattern = "chk-{name}-{global_step}-{score_name}-{score}.{ext}"
    res = _test(
        to_save,
        score_function=lambda e: e.state.score,
        score_name="acc",
        global_step_transform=lambda e, _: e.state.epoch,
        filename_pattern=pattern,
        dirname=dirname,
    )
    assert res == "chk-model-12-acc-0.9999.pt"

    pattern = "{filename_prefix}-{name}-{score}.chk"
    res = _test(to_save, "best", score_function=lambda e: e.state.score, filename_pattern=pattern, dirname=dirname)
    assert res == "best-model-0.9999.chk"

    pattern = "resnet-{filename_prefix}-{name}-{global_step}-{score}.chk"
    res = _test(
        to_save,
        "best",
        score_function=lambda e: e.state.score,
        global_step_transform=lambda e, _: e.state.epoch,
        filename_pattern=pattern,
        dirname=dirname,
    )
    assert res == "resnet-best-model-12-0.9999.chk"

    pattern = "{filename_prefix}-{name}-{score_name}-{score}.chk"
    res = _test(
        to_save,
        "best",
        score_function=lambda e: e.state.score,
        score_name="acc",
        filename_pattern=pattern,
        dirname=dirname,
    )
    assert res == "best-model-acc-0.9999.chk"

    pattern = "{global_step}-{filename_prefix}-{name}-{score_name}-{score}"
    res = _test(
        to_save,
        "best",
        score_function=lambda e: e.state.score,
        score_name="acc",
        global_step_transform=lambda e, _: e.state.epoch,
        filename_pattern=pattern,
        dirname=dirname,
    )
    assert res == "12-best-model-acc-0.9999"

    pattern = "SAVE-{name}-{score_name}-{score}.pth"
    res = _test(
        to_save,
        "best",
        score_function=lambda e: e.state.score,
        score_name="acc",
        global_step_transform=lambda e, _: e.state.epoch,
        filename_pattern=pattern,
        dirname=dirname,
    )

    assert res == "SAVE-model-acc-0.9999.pth"

    pattern = "{global_step}-chk-{filename_prefix}-{name}-{score_name}-{score}.{ext}"
    assert _test(to_save, filename_pattern=pattern, dirname=dirname) == "203-chk--model-None-None.pt"

    with pytest.raises(KeyError, match=r"random_key"):
        pattern = "SAVE-{random_key}.{ext}"
        _test(to_save, filename_pattern=pattern, dirname=dirname)


def test_setup_filename_pattern():
    # default filename pattern
    assert Checkpoint.setup_filename_pattern() == "{filename_prefix}_{name}_{global_step}_{score_name}={score}.{ext}"

    assert Checkpoint.setup_filename_pattern(False) == "{name}_{global_step}_{score_name}={score}.{ext}"
    assert Checkpoint.setup_filename_pattern(False, False, False) == "{name}_{global_step}.{ext}"
    assert Checkpoint.setup_filename_pattern(False, True, False) == "{name}_{global_step}_{score}.{ext}"
    assert Checkpoint.setup_filename_pattern(False, True, False, False) == "{name}_{score}.{ext}"
    assert Checkpoint.setup_filename_pattern(False, True, True, False) == "{name}_{score_name}={score}.{ext}"

    with pytest.raises(ValueError, match=r"At least one of with_score and with_global_step should be True."):
        Checkpoint.setup_filename_pattern(False, False, False, False)

    with pytest.raises(ValueError, match=r"If with_score_name is True, with_score should be also True"):
        Checkpoint.setup_filename_pattern(True, False, True, True)


def _setup_checkpoint():
    save_handler = MagicMock(spec=BaseSaveHandler)
    model = DummyModel()
    to_save = {"model": model}

    checkpointer = Checkpoint(to_save, save_handler=save_handler, n_saved=None)
    assert checkpointer.last_checkpoint is None

    trainer = Engine(lambda e, b: None)
    trainer.state = State(epoch=0, iteration=0)

    checkpointer(trainer)
    trainer.state.iteration = 10
    checkpointer(trainer)
    trainer.state.iteration = 20
    checkpointer(trainer)
    assert save_handler.call_count == 3
    return checkpointer


def test_checkpoint_state_dict():
    checkpointer = _setup_checkpoint()
    sd = checkpointer.state_dict()
    assert "_saved" in sd
    assert isinstance(sd["_saved"], list) and len(sd["_saved"]) == len(checkpointer._saved)

    for saved_item, true_item in zip(sd["_saved"], checkpointer._saved):
        assert saved_item[0] == true_item.priority
        assert saved_item[1] == true_item.filename


def test_checkpoint_load_state_dict():
    true_checkpointer = _setup_checkpoint()

    save_handler = MagicMock(spec=BaseSaveHandler)
    model = DummyModel()
    to_save = {"model": model}
    checkpointer = Checkpoint(to_save, save_handler=save_handler, n_saved=None)

    sd = {"_saved": [(0, "model_0.pt"), (10, "model_10.pt"), (20, "model_20.pt")]}
    checkpointer.load_state_dict(sd)
    assert checkpointer._saved == true_checkpointer._saved


@pytest.mark.parametrize(
    "to_save",
    [
        {"model": DummyModel()},
        {"model": [DummyModel(), DummyModel()]},
        {"model": {"a": {"b": DummyModel()}}},
    ],
)
def test_checkpoint__setup_checkpoint(to_save):
    save_handler = MagicMock(spec=BaseSaveHandler)
    checkpointer = Checkpoint(to_save, save_handler=save_handler, n_saved=2)
    checkpoint = checkpointer._setup_checkpoint()

    assert isinstance(checkpoint, dict)
    for k, obj in to_save.items():
        assert k in checkpoint
        if isinstance(obj, torch.nn.Module):
            assert checkpoint[k] == obj.state_dict()
        elif isinstance(obj, list):
            for c2, obj2 in zip(checkpoint[k], obj):
                assert c2 == obj2.state_dict()
        elif isinstance(obj, dict):
            c2 = checkpoint[k]
            for k2, obj2 in obj.items():
                if isinstance(obj2, torch.nn.Module):
                    assert c2[k2] == obj2.state_dict()
                elif isinstance(obj2, dict):
                    c3 = c2[k2]
                    for k3, obj3 in obj2.items():
                        assert c3[k3] == obj3.state_dict()


def test_checkpoint_fixed_filename():
    model = DummyModel()
    to_save = {"model": model}

    def _test(n_saved):
        save_handler = MagicMock(spec=BaseSaveHandler)
        checkpointer = Checkpoint(to_save, save_handler=save_handler, n_saved=n_saved, filename_pattern="{name}.{ext}")

        trainer = Engine(lambda e, b: None)

        for i in range(10):
            trainer.state = State(epoch=i, iteration=i)
            checkpointer(trainer)
            assert save_handler.call_count == i + 1
            metadata = {"basename": "model", "score_name": None, "priority": i}
            save_handler.assert_called_with(model.state_dict(), "model.pt", metadata)

    _test(None)
    _test(1)
    _test(3)


def test_checkpoint_reset():
    model = DummyModel()
    to_save = {"model": model}

    save_handler = MagicMock(spec=BaseSaveHandler)

    checkpointer = Checkpoint(to_save, save_handler=save_handler, n_saved=2)
    assert checkpointer.last_checkpoint is None

    trainer = Engine(lambda e, b: None)

    trainer.state = State(epoch=0, iteration=123)
    checkpointer(trainer)
    trainer.state.iteration = 234
    checkpointer(trainer)

    assert save_handler.call_count == 2
    assert checkpointer.last_checkpoint == "model_234.pt"
    assert len(checkpointer._saved) == 2
    assert sorted([item.filename for item in checkpointer._saved]) == sorted(["model_123.pt", "model_234.pt"])

    checkpointer.reset()
    assert len(checkpointer._saved) == 0

    trainer.state.iteration = 124
    checkpointer(trainer)

    assert save_handler.call_count == 3
    assert checkpointer.last_checkpoint == "model_124.pt"
    assert len(checkpointer._saved) == 1
    assert sorted([item.filename for item in checkpointer._saved]) == sorted(["model_124.pt"])


def test_checkpoint_reset_with_engine(dirname):
    name = "model"
    engine = Engine(lambda e, b: None)
    handler = ModelCheckpoint(dirname, _PREFIX, create_dir=False, n_saved=2)

    model = DummyModel()
    to_save = {"model": model}
    engine.add_event_handler(Events.EPOCH_COMPLETED, handler, to_save)
    engine.run([0, 1], max_epochs=10)

    expected = sorted([f"{_PREFIX}_{name}_{i}.pt" for i in [9 * 2, 10 * 2]])
    assert sorted(os.listdir(dirname)) == expected
    assert "PREFIX_model_20.pt" in str(handler.last_checkpoint)

    handler.reset()
    engine.state.max_epochs = None
    engine.run([0, 1], max_epochs=2)

    expected += [f"{_PREFIX}_{name}_{i}.pt" for i in [1 * 2, 2 * 2]]
    assert sorted(os.listdir(dirname)) == sorted(expected)
    assert "PREFIX_model_4.pt" in str(handler.last_checkpoint)


def test_greater_or_equal():
    scores = iter([1, 2, 2, 2])

    def score_function(_):
        return next(scores)

    class Saver:
        def __init__(self):
            self.counter = 0

        def __call__(self, c, f, m):
            if self.counter == 0:
                assert f == "model_1.pt"
            else:
                assert f == "model_2.pt"
            self.counter += 1

    handler = Saver()

    checkpointer = Checkpoint(
        to_save={"model": DummyModel()},
        save_handler=handler,
        score_function=score_function,
        n_saved=2,
        greater_or_equal=True,
    )
    trainer = Engine(lambda e, b: None)

    for _ in range(4):
        checkpointer(trainer)
    assert handler.counter == 4


def test_greater_or_equal_model_checkpoint(dirname):
    scores = iter([1, 2, 2, 2])

    def score_function(_):
        return next(scores)

    checkpointer = ModelCheckpoint(
        dirname,
        score_function=score_function,
        n_saved=2,
        greater_or_equal=True,
    )
    trainer = Engine(lambda e, b: None)

    to_save = {"model": DummyModel()}
    for i in range(4):
        checkpointer(trainer, to_save)
        if i == 0:
            assert Path(checkpointer.last_checkpoint).name == "model_1.pt"
        else:
            assert Path(checkpointer.last_checkpoint).name == "model_2.pt"


def test_get_default_score_fn():
    with pytest.raises(ValueError, match=r"Argument score_sign should be 1 or -1"):
        Checkpoint.get_default_score_fn("acc", 2.0)

    engine = Engine(lambda e, b: None)
    engine.state.metrics["acc"] = 0.9
    engine.state.metrics["loss"] = 0.123

    score_fn = Checkpoint.get_default_score_fn("acc")
    score = score_fn(engine)
    assert score == 0.9

    score_fn = Checkpoint.get_default_score_fn("loss", -1)
    score = score_fn(engine)
    assert score == -0.123


@pytest.mark.parametrize("obj_to_save", ["optim", "trainer"])
def test_load_single_object(obj_to_save, dirname):
    # Checks https://github.com/pytorch/ignite/issues/2479

    trainer = Engine(lambda e, b: None)
    if obj_to_save == "optim":
        t = torch.tensor(0.0)
        optim = torch.optim.SGD([t], lr=0.1)
        to_save = {"optim": optim}
    elif obj_to_save == "trainer":
        to_save = {"trainer": trainer}

    c = Checkpoint(to_save, save_handler=dirname)
    c(trainer)

    checkpoint_fp = dirname / c.last_checkpoint
    Checkpoint.load_objects(to_load=to_save, checkpoint=str(checkpoint_fp))


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.parametrize("atomic", [False, True])
def test_disksaver_distrib(distributed_context_single_node_gloo, dirname, local_rank, atomic):
    saver = DiskSaver(dirname, atomic, save_on_rank=1)
    mocked_saver = MagicMock(wraps=saver)

    mocked_saver(checkpoint={}, filename="test_disksaver_distrib.pt")

    if local_rank == 1:
        assert (dirname / "test_disksaver_distrib.pt").exists()

    else:
        mocked_saver._save_func.assert_not_called()



import numpy as np
import pytest
import torch
from sklearn.metrics import multilabel_confusion_matrix

import ignite.distributed as idist
from ignite.exceptions import NotComputableError
from ignite.metrics.multilabel_confusion_matrix import MultiLabelConfusionMatrix

torch.manual_seed(12)


def test_no_update():
    cm = MultiLabelConfusionMatrix(10)
    with pytest.raises(
        NotComputableError, match=r"Confusion matrix must have at least one example before it can be computed"
    ):
        cm.compute()


def test_num_classes_wrong_input():
    with pytest.raises(ValueError, match="Argument num_classes needs to be > 1"):
        MultiLabelConfusionMatrix(num_classes=1)


def test_multiclass_wrong_inputs():
    cm = MultiLabelConfusionMatrix(10)

    with pytest.raises(
        ValueError, match=r"y_pred must at least have shape \(batch_size, num_classes \(currently set to 10\), ...\)"
    ):
        cm.update((torch.rand(10), torch.randint(0, 2, size=(10, 10)).long()))

    with pytest.raises(
        ValueError, match=r"y must at least have shape \(batch_size, num_classes \(currently set to 10\), ...\)"
    ):
        cm.update((torch.rand(10, 10), torch.randint(0, 2, size=(10,)).long()))

    with pytest.raises(ValueError, match=r"y_pred and y have different batch size: 10 vs 8"):
        cm.update((torch.rand(10, 10), torch.randint(0, 2, size=(8, 10)).long()))

    with pytest.raises(ValueError, match=r"y does not have correct number of classes: 9 vs 10"):
        cm.update((torch.rand(10, 10), torch.randint(0, 2, size=(10, 9)).long()))

    with pytest.raises(ValueError, match=r"y_pred does not have correct number of classes: 3 vs 10"):
        cm.update((torch.rand(10, 3), torch.randint(0, 2, size=(10, 10)).long()))

    with pytest.raises(ValueError, match=r"y and y_pred shapes must match."):
        cm.update((torch.rand(10, 10, 2), torch.randint(0, 2, size=(10, 10)).long()))

    with pytest.raises(
        ValueError,
        match=r"y_pred must be of any type: \(torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64\)",
    ):
        cm.update((torch.rand(10, 10), torch.rand(10, 10)))

    with pytest.raises(
        ValueError, match=r"y must be of any type: \(torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64\)"
    ):
        cm.update((torch.rand(10, 10).type(torch.int32), torch.rand(10, 10)))

    with pytest.raises(ValueError, match=r"y_pred must be a binary tensor"):
        y = torch.randint(0, 2, size=(10, 10)).long()
        y_pred = torch.randint(0, 2, size=(10, 10)).long()
        y_pred[0, 0] = 2
        cm.update((y_pred, y))

    with pytest.raises(ValueError, match=r"y must be a binary tensor"):
        y = torch.randint(0, 2, size=(10, 10)).long()
        y_pred = torch.randint(0, 2, size=(10, 10)).long()
        y[0, 0] = 2
        cm.update((y_pred, y))


def get_y_true_y_pred():
    # Generate an image with labels 0 (background), 1, 2
    # 3 classes:
    y_true = np.zeros((1, 3, 30, 30), dtype=np.int64)
    y_true[0, 0, 5:17, 7:11] = 1
    y_true[0, 1, 1:11, 1:11] = 1
    y_true[0, 2, 15:25, 15:25] = 1

    y_pred = np.zeros((1, 3, 30, 30), dtype=np.int64)
    y_pred[0, 0, 0:7, 8:15] = 1
    y_pred[0, 1, 5:15, 1:11] = 1
    y_pred[0, 2, 20:30, 20:30] = 1
    return y_true, y_pred


def test_multiclass_images(available_device):
    num_classes = 3
    cm = MultiLabelConfusionMatrix(num_classes=num_classes, device=available_device)
    assert cm._device == torch.device(available_device)

    y_true, y_pred = get_y_true_y_pred()

    # Compute confusion matrix with sklearn
    sklearn_CM = multilabel_confusion_matrix(
        y_true.transpose((0, 2, 3, 1)).reshape(-1, 3), y_pred.transpose((0, 2, 3, 1)).reshape(-1, 3)
    )

    # Update metric
    output = (torch.tensor(y_pred), torch.tensor(y_true))
    cm.update(output)

    ignite_CM = cm.compute().cpu().numpy()

    assert np.all(ignite_CM == sklearn_CM)

    # Another test on batch of 2 images
    cm = MultiLabelConfusionMatrix(num_classes=num_classes, device=available_device)
    assert cm._device == torch.device(available_device)

    # Create a batch of two images:
    th_y_true1 = torch.tensor(y_true)
    th_y_true2 = torch.tensor(y_true.transpose(0, 1, 3, 2))
    th_y_true = torch.cat([th_y_true1, th_y_true2], dim=0)

    th_y_pred1 = torch.tensor(y_pred)
    th_y_pred2 = torch.tensor(y_pred.transpose(0, 1, 3, 2))
    th_y_pred = torch.cat([th_y_pred1, th_y_pred2], dim=0)

    # Update metric & compute
    output = (th_y_pred, th_y_true)
    cm.update(output)
    ignite_CM = cm.compute().cpu().numpy()

    # Compute confusion matrix with sklearn
    th_y_true = idist.all_gather(th_y_true)
    th_y_pred = idist.all_gather(th_y_pred)

    np_y_true = th_y_true.cpu().numpy().transpose((0, 2, 3, 1)).reshape(-1, 3)
    np_y_pred = th_y_pred.cpu().numpy().transpose((0, 2, 3, 1)).reshape(-1, 3)
    sklearn_CM = multilabel_confusion_matrix(np_y_true, np_y_pred)

    assert np.all(ignite_CM == sklearn_CM)


def _test_distrib_multiclass_images(device):
    def _test(metric_device):
        num_classes = 3
        cm = MultiLabelConfusionMatrix(num_classes=num_classes, device=metric_device)

        y_true, y_pred = get_y_true_y_pred()

        # Compute confusion matrix with sklearn
        sklearn_CM = multilabel_confusion_matrix(
            y_true.transpose((0, 2, 3, 1)).reshape(-1, 3), y_pred.transpose((0, 2, 3, 1)).reshape(-1, 3)
        )

        # Update metric
        output = (torch.tensor(y_pred).to(device), torch.tensor(y_true).to(device))
        cm.update(output)

        ignite_CM = cm.compute().cpu().numpy()

        assert np.all(ignite_CM == sklearn_CM)

        # Another test on batch of 2 images
        num_classes = 3
        cm = MultiLabelConfusionMatrix(num_classes=num_classes, device=metric_device)

        # Create a batch of two images:
        th_y_true1 = torch.tensor(y_true)
        th_y_true2 = torch.tensor(y_true.transpose(0, 1, 3, 2))
        th_y_true = torch.cat([th_y_true1, th_y_true2], dim=0)
        th_y_true = th_y_true.to(device)

        th_y_pred1 = torch.tensor(y_pred)
        th_y_pred2 = torch.tensor(y_pred.transpose(0, 1, 3, 2))
        th_y_pred = torch.cat([th_y_pred1, th_y_pred2], dim=0)
        th_y_pred = th_y_pred.to(device)

        # Update metric & compute
        output = (th_y_pred, th_y_true)
        cm.update(output)
        ignite_CM = cm.compute().cpu().numpy()

        # Compute confusion matrix with sklearn
        th_y_true = idist.all_gather(th_y_true)
        th_y_pred = idist.all_gather(th_y_pred)

        np_y_true = th_y_true.cpu().numpy().transpose((0, 2, 3, 1)).reshape(-1, 3)
        np_y_pred = th_y_pred.cpu().numpy().transpose((0, 2, 3, 1)).reshape(-1, 3)
        sklearn_CM = multilabel_confusion_matrix(np_y_true, np_y_pred)

        assert np.all(ignite_CM == sklearn_CM)

    _test("cpu")
    if device.type != "xla":
        _test(idist.device())


def _test_distrib_accumulator_device(device):
    metric_devices = [torch.device("cpu")]
    if device.type != "xla":
        metric_devices.append(idist.device())
    for metric_device in metric_devices:
        cm = MultiLabelConfusionMatrix(num_classes=3, device=metric_device)
        assert cm._device == metric_device
        assert (
            cm.confusion_matrix.device == metric_device
        ), f"{type(cm.confusion_matrix.device)}:{cm._num_correct.device} vs {type(metric_device)}:{metric_device}"

        y_true, y_pred = get_y_true_y_pred()
        cm.update((torch.tensor(y_pred), torch.tensor(y_true)))

        assert (
            cm.confusion_matrix.device == metric_device
        ), f"{type(cm.confusion_matrix.device)}:{cm._num_correct.device} vs {type(metric_device)}:{metric_device}"


def test_simple_2D_input(available_device):
    # Tests for 2D inputs with normalized = True and False

    num_iters = 5
    num_samples = 100
    num_classes = 10
    torch.manual_seed(0)
    for _ in range(num_iters):
        target = torch.randint(0, 2, size=(num_samples, num_classes))
        prediction = torch.randint(0, 2, size=(num_samples, num_classes))
        sklearn_CM = multilabel_confusion_matrix(target.cpu().numpy(), prediction.cpu().numpy())
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=False, device=available_device)
        assert mlcm._device == torch.device(available_device)
        mlcm.update([prediction, target])
        ignite_CM = mlcm.compute().cpu().numpy()
        assert np.all(sklearn_CM.astype(np.int64) == ignite_CM.astype(np.int64))
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=True, device=available_device)
        assert mlcm._device == torch.device(available_device)
        mlcm.update([prediction, target])
        ignite_CM_normalized = mlcm.compute().cpu().numpy()
        sklearn_CM_normalized = sklearn_CM / sklearn_CM.sum(axis=(1, 2))[:, None, None]
        assert np.allclose(sklearn_CM_normalized, ignite_CM_normalized)


def test_simple_ND_input(available_device):
    num_iters = 5
    num_samples = 100
    num_classes = 10
    torch.manual_seed(0)

    size_3d = 4
    for _ in range(num_iters):  # 3D tests
        target = torch.randint(0, 2, size=(num_samples, num_classes, size_3d))
        prediction = torch.randint(0, 2, size=(num_samples, num_classes, size_3d))
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=False, device=available_device)
        assert mlcm._device == torch.device(available_device)
        mlcm.update([prediction, target])
        ignite_CM = mlcm.compute().cpu().numpy()
        target_reshaped = target.permute(0, 2, 1).reshape(size_3d * num_samples, num_classes)
        prediction_reshaped = prediction.permute(0, 2, 1).reshape(size_3d * num_samples, num_classes)
        sklearn_CM = multilabel_confusion_matrix(target_reshaped.cpu().numpy(), prediction_reshaped.cpu().numpy())
        assert np.all(sklearn_CM.astype(np.int64) == ignite_CM.astype(np.int64))

    size_4d = 4
    for _ in range(num_iters):  # 4D tests
        target = torch.randint(0, 2, size=(num_samples, num_classes, size_3d, size_4d))
        prediction = torch.randint(0, 2, size=(num_samples, num_classes, size_3d, size_4d))
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=False, device=available_device)
        assert mlcm._device == torch.device(available_device)
        mlcm.update([prediction, target])
        ignite_CM = mlcm.compute().cpu().numpy()
        target_reshaped = target.permute(0, 2, 3, 1).reshape(size_3d * size_4d * num_samples, num_classes)
        prediction_reshaped = prediction.permute(0, 2, 3, 1).reshape(size_3d * size_4d * num_samples, num_classes)
        sklearn_CM = multilabel_confusion_matrix(target_reshaped.cpu().numpy(), prediction_reshaped.cpu().numpy())
        assert np.all(sklearn_CM.astype(np.int64) == ignite_CM.astype(np.int64))

    size_5d = 4
    for _ in range(num_iters):  # 5D tests
        target = torch.randint(0, 2, size=(num_samples, num_classes, size_3d, size_4d, size_5d))
        prediction = torch.randint(0, 2, size=(num_samples, num_classes, size_3d, size_4d, size_5d))
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=False, device=available_device)
        assert mlcm._device == torch.device(available_device)
        mlcm.update([prediction, target])
        ignite_CM = mlcm.compute().cpu().numpy()
        target_reshaped = target.permute(0, 2, 3, 4, 1).reshape(size_3d * size_4d * size_5d * num_samples, num_classes)
        prediction_reshaped = prediction.permute(0, 2, 3, 4, 1).reshape(
            size_3d * size_4d * size_5d * num_samples, num_classes
        )
        sklearn_CM = multilabel_confusion_matrix(target_reshaped.cpu().numpy(), prediction_reshaped.cpu().numpy())
        assert np.all(sklearn_CM.astype(np.int64) == ignite_CM.astype(np.int64))


def test_simple_batched(available_device):
    num_iters = 5
    num_samples = 100
    num_classes = 10
    batch_size = 1
    torch.manual_seed(0)

    for _ in range(num_iters):  # 2D tests
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=False, device=available_device)
        assert mlcm._device == torch.device(available_device)
        targets = torch.randint(0, 2, size=(int(num_samples / batch_size), batch_size, num_classes))
        predictions = torch.randint(0, 2, size=(int(num_samples / batch_size), batch_size, num_classes))
        for i in range(int(num_samples / batch_size)):
            target_sample = targets[i]
            prediction_sample = predictions[i]
            mlcm.update([prediction_sample, target_sample])

        ignite_CM = mlcm.compute().cpu().numpy()
        targets_reshaped = targets.reshape(-1, num_classes)
        predictions_reshaped = predictions.reshape(-1, num_classes)
        sklearn_CM = multilabel_confusion_matrix(targets_reshaped.cpu().numpy(), predictions_reshaped.cpu().numpy())
        assert np.all(sklearn_CM.astype(np.int64) == ignite_CM.astype(np.int64))

    size_3d = 4
    for _ in range(num_iters):  # 3D tests
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=False, device=available_device)
        assert mlcm._device == torch.device(available_device)
        targets = torch.randint(0, 2, size=(int(num_samples / batch_size), batch_size, num_classes, size_3d))
        predictions = torch.randint(0, 2, size=(int(num_samples / batch_size), batch_size, num_classes, size_3d))
        for i in range(int(num_samples / batch_size)):
            target_sample = targets[i]
            prediction_sample = predictions[i]
            mlcm.update([prediction_sample, target_sample])

        ignite_CM = mlcm.compute().cpu().numpy()
        targets_reshaped = targets.permute(0, 1, 3, 2).reshape(-1, num_classes)
        predictions_reshaped = predictions.permute(0, 1, 3, 2).reshape(-1, num_classes)
        sklearn_CM = multilabel_confusion_matrix(targets_reshaped.cpu().numpy(), predictions_reshaped.cpu().numpy())
        assert np.all(sklearn_CM.astype(np.int64) == ignite_CM.astype(np.int64))

    size_4d = 4
    for _ in range(num_iters):  # 4D tests
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=False, device=available_device)
        assert mlcm._device == torch.device(available_device)
        targets = torch.randint(0, 2, size=(int(num_samples / batch_size), batch_size, num_classes, size_3d, size_4d))
        predictions = torch.randint(
            0, 2, size=(int(num_samples / batch_size), batch_size, num_classes, size_3d, size_4d)
        )
        for i in range(int(num_samples / batch_size)):
            target_sample = targets[i]
            prediction_sample = predictions[i]
            mlcm.update([prediction_sample, target_sample])

        ignite_CM = mlcm.compute().cpu().numpy()
        targets_reshaped = targets.permute(0, 1, 3, 4, 2).reshape(-1, num_classes)
        predictions_reshaped = predictions.permute(0, 1, 3, 4, 2).reshape(-1, num_classes)
        sklearn_CM = multilabel_confusion_matrix(targets_reshaped.cpu().numpy(), predictions_reshaped.cpu().numpy())
        assert np.all(sklearn_CM.astype(np.int64) == ignite_CM.astype(np.int64))

    size_5d = 4
    for _ in range(num_iters):  # 5D tests
        mlcm = MultiLabelConfusionMatrix(num_classes, normalized=False, device=available_device)
        assert mlcm._device == torch.device(available_device)
        targets = torch.randint(
            0, 2, size=(int(num_samples / batch_size), batch_size, num_classes, size_3d, size_4d, size_5d)
        )
        predictions = torch.randint(
            0, 2, size=(int(num_samples / batch_size), batch_size, num_classes, size_3d, size_4d, size_5d)
        )
        for i in range(int(num_samples / batch_size)):
            target_sample = targets[i]
            prediction_sample = predictions[i]
            mlcm.update([prediction_sample, target_sample])

        ignite_CM = mlcm.compute().cpu().numpy()
        targets_reshaped = targets.permute(0, 1, 3, 4, 5, 2).reshape(-1, num_classes)
        predictions_reshaped = predictions.permute(0, 1, 3, 4, 5, 2).reshape(-1, num_classes)
        sklearn_CM = multilabel_confusion_matrix(targets_reshaped.cpu().numpy(), predictions_reshaped.cpu().numpy())
        assert np.all(sklearn_CM.astype(np.int64) == ignite_CM.astype(np.int64))


# @pytest.mark.distributed
# @pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
# @pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
# def test_distrib_nccl_gpu(distributed_context_single_node_nccl):

#     device = idist.device()
#     _test_distrib_multiclass_images(device)
#     _test_distrib_accumulator_device(device)


# @pytest.mark.distributed
# @pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
# def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):

#     device = idist.device()
#     _test_distrib_multiclass_images(device)
#     _test_distrib_accumulator_device(device)


# @pytest.mark.distributed
# @pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
# @pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
# def test_distrib_hvd(gloo_hvd_executor):

#     device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
#     nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

#     gloo_hvd_executor(_test_distrib_multiclass_images, (device,), np=nproc, do_init=True)
#     gloo_hvd_executor(_test_distrib_accumulator_device, (device,), np=nproc, do_init=True)


# @pytest.mark.multinode_distributed
# @pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
# @pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
# def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
#
#     device = idist.device()
#     _test_distrib_multiclass_images(device)
#     _test_distrib_accumulator_device(device)


# @pytest.mark.multinode_distributed
# @pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
# @pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
# def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
#
#     device = idist.device()
#     _test_distrib_multiclass_images(device)
#     _test_distrib_accumulator_device(device)


# @pytest.mark.tpu
# @pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
# @pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
# def test_distrib_single_device_xla():
#     device = idist.device()
#     _test_distrib_multiclass_images(device)
#     _test_distrib_accumulator_device(device)


# def _test_distrib_xla_nprocs(index):
#     device = idist.device()
#     _test_distrib_multiclass_images(device)
#     _test_distrib_accumulator_device(device)


# @pytest.mark.tpu
# @pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
# @pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
# def test_distrib_xla_nprocs(xmp_executor):
#     n = int(os.environ["NUM_TPU_WORKERS"])
#     xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



from typing import Tuple

import pytest

import torch
from scipy.stats import spearmanr
from torch import Tensor

from ignite import distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics.regression import SpearmanRankCorrelation


def test_zero_sample():
    with pytest.raises(
        NotComputableError, match="SpearmanRankCorrelation must have at least one example before it can be computed"
    ):
        metric = SpearmanRankCorrelation()
        metric.compute()


def test_wrong_y_pred_shape():
    with pytest.raises(ValueError, match=r"Input y_pred should have shape \(N,\) or \(N, 1\), but given"):
        metric = SpearmanRankCorrelation()
        y_pred = torch.arange(9).reshape(3, 3).float()
        y = torch.arange(3).unsqueeze(1).float()
        metric.update((y_pred, y))


def test_wrong_y_shape():
    with pytest.raises(ValueError, match=r"Input y should have shape \(N,\) or \(N, 1\), but given"):
        metric = SpearmanRankCorrelation()
        y_pred = torch.arange(3).unsqueeze(1).float()
        y = torch.arange(9).reshape(3, 3).float()
        metric.update((y_pred, y))


def test_wrong_y_pred_dtype():
    with pytest.raises(TypeError, match="Input y_pred dtype should be float 16, 32 or 64, but given"):
        metric = SpearmanRankCorrelation()
        y_pred = torch.arange(3).unsqueeze(1).long()
        y = torch.arange(3).unsqueeze(1).float()
        metric.update((y_pred, y))


def test_wrong_y_dtype():
    with pytest.raises(TypeError, match="Input y dtype should be float 16, 32 or 64, but given"):
        metric = SpearmanRankCorrelation()
        y_pred = torch.arange(3).unsqueeze(1).float()
        y = torch.arange(3).unsqueeze(1).long()
        metric.update((y_pred, y))


def test_spearman_correlation(available_device):
    torch.manual_seed(0)

    inputs = [torch.randn(4) for _ in range(4)]
    ground_truth = torch.randn(4)

    m = SpearmanRankCorrelation(device=available_device)
    assert m._device == torch.device(available_device)

    all_preds = []
    all_targets = []

    for x in inputs:
        m.update((x, ground_truth))
        all_preds.append(x)
        all_targets.append(ground_truth)

        pred_cat = torch.cat(all_preds).numpy()
        target_cat = torch.cat(all_targets).numpy()
        expected = spearmanr(pred_cat, target_cat).statistic
        assert m.compute() == pytest.approx(expected, rel=1e-4)


@pytest.fixture(params=list(range(2)))
def test_case(request):
    # correlated sample
    x = torch.randn(size=[50]).float()
    y = x + torch.randn_like(x) * 0.1

    return [
        (x, y, 1),
        (torch.rand(size=(50, 1)).float(), torch.rand(size=(50, 1)).float(), 10),
    ][request.param]


@pytest.mark.parametrize("n_times", range(5))
def test_integration_spearman_correlation(n_times, test_case: Tuple[Tensor, Tensor, int], available_device):
    y_pred, y, batch_size = test_case

    def update_fn(engine: Engine, batch):
        idx = (engine.state.iteration - 1) * batch_size
        y_true_batch = y[idx : idx + batch_size]
        y_pred_batch = y_pred[idx : idx + batch_size]
        return y_pred_batch, y_true_batch

    engine = Engine(update_fn)

    m = SpearmanRankCorrelation(device=available_device)
    assert m._device == torch.device(available_device)
    m.attach(engine, "spearman_corr")

    data = list(range(y_pred.shape[0] // batch_size))
    corr = engine.run(data, max_epochs=1).metrics["spearman_corr"]

    # Convert only for computing the expected value
    expected = spearmanr(y_pred.numpy().ravel(), y.numpy().ravel()).statistic

    assert pytest.approx(expected, rel=2e-4) == corr


@pytest.mark.usefixtures("distributed")
class TestDistributed:
    def test_compute(self):
        rank = idist.get_rank()
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)

        torch.manual_seed(10 + rank)
        for metric_device in metric_devices:
            m = SpearmanRankCorrelation(device=metric_device)

            y_pred = torch.rand(size=[100], device=device)
            y = torch.rand(size=[100], device=device)

            m.update((y_pred, y))

            y_pred = idist.all_gather(y_pred)
            y = idist.all_gather(y)

            np_y = y.cpu().numpy()
            np_y_pred = y_pred.cpu().numpy()

            np_ans = spearmanr(np_y_pred, np_y).statistic

            assert pytest.approx(np_ans, rel=2e-4) == m.compute()

    @pytest.mark.parametrize("n_epochs", [1, 2])
    def test_integration(self, n_epochs: int):
        tol = 2e-4
        rank = idist.get_rank()
        device = idist.device()
        metric_devices = [torch.device("cpu")]
        if device.type != "xla":
            metric_devices.append(device)

        n_iters = 80
        batch_size = 16

        for metric_device in metric_devices:
            torch.manual_seed(12 + rank)

            y_true = torch.rand(size=(n_iters * batch_size,)).to(device)
            y_preds = torch.rand(size=(n_iters * batch_size,)).to(device)

            engine = Engine(
                lambda e, i: (
                    y_preds[i * batch_size : (i + 1) * batch_size],
                    y_true[i * batch_size : (i + 1) * batch_size],
                )
            )

            corr = SpearmanRankCorrelation(device=metric_device)
            corr.attach(engine, "spearman_corr")

            data = list(range(n_iters))
            engine.run(data=data, max_epochs=n_epochs)

            y_preds = idist.all_gather(y_preds)
            y_true = idist.all_gather(y_true)

            assert "spearman_corr" in engine.state.metrics

            res = engine.state.metrics["spearman_corr"]

            np_y = y_true.cpu().numpy()
            np_y_pred = y_preds.cpu().numpy()

            np_ans = spearmanr(np_y_pred, np_y).statistic

            assert pytest.approx(np_ans, rel=tol) == res



import torch

import ignite.distributed as idist
from ignite.distributed.comp_models.base import _torch_version_gt_112
from tests.ignite.distributed.utils import (
    _sanity_check,
    _test_distrib__get_max_length,
    _test_distrib_all_gather,
    _test_distrib_all_reduce,
    _test_distrib_barrier,
    _test_distrib_broadcast,
    _test_distrib_group,
    _test_idist_all_gather_tensors_with_shapes,
    _test_sync,
)


def test_no_distrib(capsys):
    assert idist.backend() is None
    if torch.cuda.is_available():
        assert idist.device().type == "cuda"
    elif _torch_version_gt_112 and torch.backends.mps.is_available():
        assert idist.device().type == "mps"
    else:
        assert idist.device().type == "cpu"
    assert idist.get_rank() == 0
    assert idist.get_world_size() == 1
    assert idist.get_local_rank() == 0
    assert idist.model_name() == "serial"

    from ignite.distributed.utils import _model, _SerialModel

    _sanity_check()
    assert isinstance(_model, _SerialModel)

    idist.show_config()
    captured = capsys.readouterr()
    out = captured.err.split("\r")
    out = list(map(lambda x: x.strip(), out))
    out = list(filter(None, out))
    assert "ignite.distributed.utils INFO: distributed configuration: serial" in out[-1]
    assert "ignite.distributed.utils INFO: backend: None" in out[-1]
    if torch.cuda.is_available():
        assert "ignite.distributed.utils INFO: device: cuda" in out[-1]
    elif _torch_version_gt_112 and torch.backends.mps.is_available():
        assert "ignite.distributed.utils INFO: device: mps" in out[-1]
    else:
        assert "ignite.distributed.utils INFO: device: cpu" in out[-1]
    assert "ignite.distributed.utils INFO: rank: 0" in out[-1]
    assert "ignite.distributed.utils INFO: local rank: 0" in out[-1]
    assert "ignite.distributed.utils INFO: world size: 1" in out[-1]


def test_sync_no_dist():
    from ignite.distributed.comp_models import _SerialModel

    _test_sync(_SerialModel)


def test_idist_methods_no_dist():
    assert idist.get_world_size() < 2
    assert idist.backend() is None, f"{idist.backend()}"


def test_idist__model_methods_no_dist():
    _test_distrib__get_max_length("cpu")
    if torch.cuda.device_count() > 1:
        _test_distrib__get_max_length("cuda")


def test_idist_collective_ops_no_dist():
    _test_distrib_all_reduce("cpu")
    _test_distrib_all_gather("cpu")
    _test_idist_all_gather_tensors_with_shapes("cpu")
    _test_distrib_barrier("cpu")
    _test_distrib_broadcast("cpu")
    _test_distrib_group("cpu")

    if torch.cuda.device_count() > 1:
        _test_distrib_all_reduce("cuda")
        _test_distrib_all_gather("cuda")
        _test_idist_all_gather_tensors_with_shapes("cuda")
        _test_distrib_barrier("cuda")
        _test_distrib_broadcast("cuda")
        _test_distrib_group("cuda")



import sys


class Progbar(object):
    def __init__(self, loader, metrics):
        self.num_iterations = len(loader)
        self.output_stream = sys.stdout
        self.metrics = metrics
        self.alpha = 0.98

    def _calc_running_avg(self, engine):
        for k, v in engine.state.output.items():
            old_v = self.metrics.get(k, v)
            new_v = self.alpha * old_v + (1 - self.alpha) * v
            self.metrics[k] = new_v

    def __call__(self, engine):
        self._calc_running_avg(engine)
        num_seen = engine.state.iteration - self.num_iterations * (engine.state.epoch - 1)

        percent_seen = 100 * float(num_seen) / self.num_iterations
        equal_to = int(percent_seen / 10)
        done = int(percent_seen) == 100

        bar = "[" + "=" * equal_to + ">" * (not done) + " " * (10 - equal_to) + "]"
        message = f"Epoch {engine.state.epoch} | {percent_seen:.2f}% | {bar}"
        for key, value in self.metrics.items():
            message += f" | {key}: {value:.2e}"

        message += "\r"

        self.output_stream.write(message)
        self.output_stream.flush()

        if done:
            self.output_stream.write("\n")



from collections import Counter
from typing import Any, Sequence, Tuple

__all__ = ["ngrams", "lcs", "modified_precision"]


def ngrams(sequence: Sequence[Any], n: int) -> Counter:
    """
    Generate the ngrams from a sequence of items

    Args:
        sequence: sequence of items
        n: n-gram order

    Returns:
        A counter of ngram objects

    .. versionadded:: 0.4.5
    """
    return Counter([tuple(sequence[i : i + n]) for i in range(len(sequence) - n + 1)])


def lcs(seq_a: Sequence[Any], seq_b: Sequence[Any]) -> int:
    """
    Compute the length of the longest common subsequence in two sequence of items
    https://en.wikipedia.org/wiki/Longest_common_subsequence_problem

    Args:
        seq_a: first sequence of items
        seq_b: second sequence of items

    Returns:
        The length of the longest common subsequence

    .. versionadded:: 0.4.5
    """
    m = len(seq_a)
    n = len(seq_b)

    dp = [[0] * (n + 1) for _ in range(m + 1)]

    for i in range(m + 1):
        for j in range(n + 1):
            if i == 0 or j == 0:
                dp[i][j] = 0
            elif seq_a[i - 1] == seq_b[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
            else:
                dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])

    return dp[m][n]


def modified_precision(references: Sequence[Sequence[Any]], candidate: Any, n: int) -> Tuple[int, int]:
    """
    Compute the modified precision

    .. math::
       p_{n} = \frac{m_{n}}{l_{n}}

    where m_{n} is the number of matched n-grams between translation T and its reference R, and l_{n} is the
    total number of n-grams in the translation T.

    More details can be found in `Papineni et al. 2002`__.

    __ https://aclanthology.org/P02-1040

    Args:
        references: list of references R
        candidate: translation T
        n: n-gram order

    Returns:
        The length of the longest common subsequence

    .. versionadded:: 0.4.5
    """
    # ngrams of the candidate
    counts = ngrams(candidate, n)

    # union of ngrams of references
    max_counts: Counter = Counter()
    for reference in references:
        max_counts |= ngrams(reference, n)

    # clipped count of the candidate and references
    clipped_counts = counts & max_counts

    return sum(clipped_counts.values()), sum(counts.values())



import os

import numpy as np
import pytest
import torch
from pytest import approx, raises

import ignite.distributed as idist
from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics.regression import MeanAbsoluteRelativeError


def test_wrong_input_shapes():
    m = MeanAbsoluteRelativeError()

    with raises(ValueError, match=r"Input data shapes should be the same, but given"):
        m.update((torch.rand(4), torch.rand(4, 1)))

    with raises(ValueError, match=r"Input data shapes should be the same, but given"):
        m.update((torch.rand(4, 1), torch.rand(4)))


def test_mean_absolute_relative_error(available_device):
    a = torch.rand(4)
    b = torch.rand(4)
    c = torch.rand(4)
    d = torch.rand(4)
    ground_truth = torch.rand(4)

    m = MeanAbsoluteRelativeError(device=available_device)
    assert m._device == torch.device(available_device)

    m.update((a, ground_truth))
    abs_error_a = torch.sum(torch.abs(ground_truth - a) / torch.abs(ground_truth))
    num_samples_a = a.size()[0]
    sum_error = abs_error_a
    sum_samples = num_samples_a
    MARE_a = sum_error / sum_samples
    assert m.compute() == approx(MARE_a.item())

    m.update((b, ground_truth))
    abs_error_b = torch.sum(torch.abs(ground_truth - b) / torch.abs(ground_truth))
    num_samples_b = b.size()[0]
    sum_error += abs_error_b
    sum_samples += num_samples_b
    MARE_b = sum_error / sum_samples
    assert m.compute() == approx(MARE_b.item())

    m.update((c, ground_truth))
    abs_error_c = torch.sum(torch.abs(ground_truth - c) / torch.abs(ground_truth))
    num_samples_c = c.size()[0]
    sum_error += abs_error_c
    sum_samples += num_samples_c
    MARE_c = sum_error / sum_samples
    assert m.compute() == approx(MARE_c.item())

    m.update((d, ground_truth))
    abs_error_d = torch.sum(torch.abs(ground_truth - d) / torch.abs(ground_truth))
    num_samples_d = d.size()[0]
    sum_error += abs_error_d
    sum_samples += num_samples_d
    MARE_d = sum_error / sum_samples
    assert m.compute() == approx(MARE_d.item())


def test_zero_div():
    a = torch.tensor([2.0, -1.0, -1.0, 2.0])
    ground_truth = torch.tensor([0.0, 0.5, 0.2, 1.0])

    m = MeanAbsoluteRelativeError()
    with raises(NotComputableError, match=r"The ground truth has 0"):
        m.update((a, ground_truth))


def test_zero_sample():
    m = MeanAbsoluteRelativeError()
    with raises(NotComputableError, match=r"MeanAbsoluteRelativeError must have at least one sample"):
        m.compute()


@pytest.mark.parametrize("n_times", range(5))
@pytest.mark.parametrize(
    "test_cases",
    [
        (torch.rand(size=(100,)), torch.rand(size=(100,)), 10),
        (torch.rand(size=(100, 1)), torch.rand(size=(100, 1)), 20),
    ],
)
def test_integration_mean_absolute_relative_error(n_times, test_cases, available_device):
    y_pred, y, batch_size = test_cases

    def update_fn(engine, batch):
        idx = (engine.state.iteration - 1) * batch_size
        y_true_batch = y[idx : idx + batch_size]
        y_pred_batch = y_pred[idx : idx + batch_size]
        return y_pred_batch, y_true_batch

    engine = Engine(update_fn)

    m = MeanAbsoluteRelativeError(device=available_device)
    assert m._device == torch.device(available_device)
    m.attach(engine, "mare")

    data = list(range(y_pred.shape[0] // batch_size))
    mare = engine.run(data, max_epochs=1).metrics["mare"]

    eps = 1e-20  # to avoid division by zero
    res = torch.abs(y_pred - y) / (torch.abs(y) + eps)
    expected = res.mean().item()

    assert mare == pytest.approx(expected)


def _test_distrib_compute(device):
    rank = idist.get_rank()

    def _test(metric_device):
        metric_device = torch.device(metric_device)
        m = MeanAbsoluteRelativeError(device=metric_device)

        y_pred = torch.randint(1, 11, size=(10,), device=device).float()
        y = torch.randint(1, 11, size=(10,), device=device).float()

        m.update((y_pred, y))

        # gather y_pred, y
        y_pred = idist.all_gather(y_pred)
        y = idist.all_gather(y)

        np_y_pred = y_pred.cpu().numpy()
        np_y = y.cpu().numpy()

        res = m.compute()

        abs_error = np.sum(abs(np_y - np_y_pred) / abs(np_y))
        num_samples = len(y_pred)
        np_res = abs_error / num_samples

        assert np_res == approx(res)

    for i in range(3):
        torch.manual_seed(10 + rank + i)
        _test("cpu")
        if device.type != "xla":
            _test(idist.device())


def _test_distrib_integration(device):
    rank = idist.get_rank()

    def _test(n_epochs, metric_device):
        metric_device = torch.device(metric_device)
        n_iters = 80
        batch_size = 16

        y_true = torch.rand(size=(n_iters * batch_size,)).to(device)
        y_preds = torch.rand(size=(n_iters * batch_size,)).to(device)

        def update(engine, i):
            return (
                y_preds[i * batch_size : (i + 1) * batch_size],
                y_true[i * batch_size : (i + 1) * batch_size],
            )

        engine = Engine(update)

        m = MeanAbsoluteRelativeError(device=metric_device)
        m.attach(engine, "mare")

        data = list(range(n_iters))
        engine.run(data=data, max_epochs=n_epochs)

        y_preds = idist.all_gather(y_preds)
        y_true = idist.all_gather(y_true)

        assert "mare" in engine.state.metrics

        mare = engine.state.metrics["mare"]

        np_y_true = y_true.cpu().numpy()
        np_y_preds = y_preds.cpu().numpy()

        abs_error = np.sum(abs(np_y_true - np_y_preds) / abs(np_y_true))
        num_samples = len(y_preds)
        np_res = abs_error / num_samples

        assert approx(mare) == np_res

    metric_devices = ["cpu"]
    if device.type != "xla":
        metric_devices.append(idist.device())
    for metric_device in metric_devices:
        for i in range(2):
            torch.manual_seed(12 + rank + i)
            _test(n_epochs=1, metric_device=metric_device)
            _test(n_epochs=2, metric_device=metric_device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif(torch.cuda.device_count() < 1, reason="Skip if no GPU")
def test_distrib_nccl_gpu(distributed_context_single_node_nccl):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
def test_distrib_gloo_cpu_or_gpu(distributed_context_single_node_gloo):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.distributed
@pytest.mark.skipif(not idist.has_hvd_support, reason="Skip if no Horovod dist support")
@pytest.mark.skipif("WORLD_SIZE" in os.environ, reason="Skip if launched as multiproc")
def test_distrib_hvd(gloo_hvd_executor):
    device = torch.device("cpu" if not torch.cuda.is_available() else "cuda")
    nproc = 4 if not torch.cuda.is_available() else torch.cuda.device_count()

    gloo_hvd_executor(_test_distrib_compute, (device,), np=nproc, do_init=True)
    gloo_hvd_executor(_test_distrib_integration, (device,), np=nproc, do_init=True)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_gloo_cpu_or_gpu(distributed_context_multi_node_gloo):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.multinode_distributed
@pytest.mark.skipif(not idist.has_native_dist_support, reason="Skip if no native dist support")
@pytest.mark.skipif("GPU_MULTINODE_DISTRIB" not in os.environ, reason="Skip if not multi-node distributed")
def test_multinode_distrib_nccl_gpu(distributed_context_multi_node_nccl):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" in os.environ, reason="Skip if NUM_TPU_WORKERS is in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_single_device_xla():
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


def _test_distrib_xla_nprocs(index):
    device = idist.device()
    _test_distrib_compute(device)
    _test_distrib_integration(device)


@pytest.mark.tpu
@pytest.mark.skipif("NUM_TPU_WORKERS" not in os.environ, reason="Skip if no NUM_TPU_WORKERS in env vars")
@pytest.mark.skipif(not idist.has_xla_support, reason="Skip if no PyTorch XLA package")
def test_distrib_xla_nprocs(xmp_executor):
    n = int(os.environ["NUM_TPU_WORKERS"])
    xmp_executor(_test_distrib_xla_nprocs, args=(), nprocs=n)



# -*- coding: utf-8 -*-
import sys
import time
from argparse import Namespace
from unittest.mock import patch

import numpy as np
import pytest
import torch
from packaging.version import Version

from ignite.engine import Engine, Events

from ignite.handlers import ProgressBar, TerminateOnNan
from ignite.metrics import RunningAverage

if sys.platform.startswith("win"):
    pytest.skip("Skip on Windows", allow_module_level=True)


def get_tqdm_version():
    import tqdm

    return Version(tqdm.__version__)


def update_fn(engine, batch):
    a = 1
    engine.state.metrics["a"] = a
    return a


def test_pbar_errors():
    with pytest.raises(ModuleNotFoundError, match=r"This contrib module requires tqdm to be installed"):
        with patch.dict("sys.modules", {"tqdm.autonotebook": None}):
            ProgressBar(ncols=80)

    pbar = ProgressBar(ncols=80)
    with pytest.raises(ValueError, match=r"Logging event abc is not in allowed"):
        pbar.attach(Engine(lambda e, b: None), event_name=Namespace(name="abc"))


def test_pbar(capsys):
    n_epochs = 2
    loader = [1, 2]
    engine = Engine(update_fn)

    pbar = ProgressBar(ncols=80)
    pbar.attach(engine, ["a"])

    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    if get_tqdm_version() < Version("4.49.0"):
        expected = "Epoch 8 -*-     , a=1 [00:00<00:00]"
    else:
        expected = "Epoch [2/2]: [1/2]  50%|████████████████████▌                    , a=1 [00:00<?]"
    assert err[-1] == expected


def test_pbar_file(tmp_path):
    n_epochs = 2
    loader = [1, 2]
    engine = Engine(update_fn)

    file_path = tmp_path / "temp.txt"
    file = open(str(file_path), "w+")

    pbar = ProgressBar(file=file, ncols=80)
    pbar.attach(engine, ["a"])
    engine.run(loader, max_epochs=n_epochs)

    file.close()  # Force a flush of the buffer. file.flush() does not work.

    file = open(str(file_path), "r")
    lines = file.readlines()

    if get_tqdm_version() < Version("4.49.0"):
        expected = "Epoch [2/2]: [1/2]  50%|█████     , a=1 [00:00<00:00]\n"
    else:
        expected = "Epoch [2/2]: [1/2]  50%|████████████████████▌                    , a=1 [00:00<?]\n"
    assert lines[-2] == expected


def test_pbar_log_message(capsys):
    pbar = ProgressBar(ncols=80)

    pbar.log_message("test")

    captured = capsys.readouterr()
    out = captured.out.split("\r")
    out = list(map(lambda x: x.strip(), out))
    out = list(filter(None, out))
    expected = "test"
    assert out[-1] == expected


def test_pbar_log_message_file(tmp_path):
    file_path = tmp_path / "temp.txt"
    file = open(str(file_path), "w+")

    pbar = ProgressBar(file=file, ncols=80)
    pbar.log_message("test")

    file.close()  # Force a flush of the buffer. file.flush() does not work.

    file = open(str(file_path), "r")
    lines = file.readlines()

    expected = "test\n"
    assert lines[0] == expected


def test_attach_fail_with_string():
    engine = Engine(update_fn)
    pbar = ProgressBar(ncols=80)

    with pytest.raises(TypeError):
        pbar.attach(engine, "a")


def test_pbar_batch_indeces(capsys):
    engine = Engine(lambda e, b: time.sleep(0.1))

    @engine.on(Events.ITERATION_STARTED)
    def print_iter(_):
        print("iteration: ", engine.state.iteration)

    ProgressBar(persist=True, ncols=80).attach(engine)
    engine.run(list(range(4)), max_epochs=1)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    printed_batch_indeces = set(map(lambda x: int(x.split("/")[0][-1]), err))
    expected_batch_indeces = list(range(1, 5))
    assert sorted(list(printed_batch_indeces)) == expected_batch_indeces


def test_pbar_with_metric(capsys):
    n_iters = 2
    data = list(range(n_iters))
    loss_values = iter(range(n_iters))

    def step(engine, batch):
        loss_value = next(loss_values)
        return loss_value

    trainer = Engine(step)

    RunningAverage(alpha=0.5, output_transform=lambda x: x).attach(trainer, "batchloss")

    pbar = ProgressBar(ncols=80)
    pbar.attach(trainer, metric_names=["batchloss"])

    trainer.run(data=data, max_epochs=1)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    actual = err[-1]
    if get_tqdm_version() < Version("4.49.0"):
        expected = "Iteration: [1/2]  50%|██████     , batchloss=0.5 [00:00<00:00]"
    else:
        expected = "Iteration: [1/2]  50%|████████████████▌                , batchloss=0.5 [00:00<?]"
    assert actual == expected


def test_pbar_with_all_metric(capsys):
    n_iters = 2
    data = list(range(n_iters))
    loss_values = iter(range(n_iters))
    another_loss_values = iter(range(1, n_iters + 1))

    def step(engine, batch):
        loss_value = next(loss_values)
        another_loss_value = next(another_loss_values)
        return loss_value, another_loss_value

    trainer = Engine(step)

    RunningAverage(alpha=0.5, output_transform=lambda x: x[0]).attach(trainer, "batchloss")
    RunningAverage(alpha=0.5, output_transform=lambda x: x[1]).attach(trainer, "another batchloss")

    pbar = ProgressBar(ncols=80)
    pbar.attach(trainer, metric_names="all")

    trainer.run(data=data, max_epochs=1)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    actual = err[-1]
    if get_tqdm_version() < Version("4.49.0"):
        expected = "Iteration: [1/2]  50%|███   , batchloss=0.5, another batchloss=1.5 [00:00<00:00]"
    else:
        expected = "Iteration: [1/2]  50%|█████     , batchloss=0.5, another batchloss=1.5 [00:00<?]"
    assert actual == expected


def test_pbar_with_state_attrs(capsys):
    n_iters = 2
    data = list(range(n_iters))
    loss_values = iter(range(n_iters))

    def step(engine, batch):
        loss_value = next(loss_values)
        return loss_value

    trainer = Engine(step)
    trainer.state.alpha = 3.899
    trainer.state.beta = torch.tensor(12.21)
    trainer.state.gamma = torch.tensor([21.0, 6.0])

    RunningAverage(alpha=0.5, output_transform=lambda x: x).attach(trainer, "batchloss")

    pbar = ProgressBar(ncols=80)
    pbar.attach(trainer, metric_names=["batchloss"], state_attributes=["alpha", "beta", "gamma"])

    trainer.run(data=data, max_epochs=1)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    actual = err[-1]
    if get_tqdm_version() < Version("4.49.0"):
        expected = (
            "Iteration: [1/2]  50%|█████     , batchloss=0.5, alpha=3.9, beta=12.2, gamma_0=21, gamma_1=6 [00:00<00:00]"
        )
    else:
        expected = "Iteration: [1/2]  50%|▌, batchloss=0.5, alpha=3.9, beta=12.2, gamma_0=21, gamma_"
    assert actual == expected


def test_pbar_no_metric_names(capsys):
    n_epochs = 2
    loader = [1, 2]
    engine = Engine(update_fn)

    pbar = ProgressBar(ncols=80)
    pbar.attach(engine)

    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    actual = err[-1]
    if get_tqdm_version() < Version("4.49.0"):
        expected = "Epoch [2/2]: [1/2]  50%|██████████            [00:00<00:00]"
    else:
        expected = "Epoch [2/2]: [1/2]  50%|███████████████████████                        [00:00<?]"
    assert actual == expected


def test_pbar_with_output(capsys):
    n_epochs = 2
    loader = [1, 2]
    engine = Engine(update_fn)

    pbar = ProgressBar(ncols=80)
    pbar.attach(engine, output_transform=lambda x: {"a": x})

    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    if get_tqdm_version() < Version("4.49.0"):
        expected = "Epoch [2/2]: [1/2]  50%|█████     , a=1 [00:00<00:00]"
    else:
        expected = "Epoch [2/2]: [1/2]  50%|████████████████████▌                    , a=1 [00:00<?]"
    assert err[-1] == expected


def test_pbar_fail_with_non_callable_transform():
    engine = Engine(update_fn)
    pbar = ProgressBar(ncols=80)

    with pytest.raises(TypeError):
        pbar.attach(engine, output_transform=1)


def test_pbar_with_scalar_output(capsys):
    n_epochs = 2
    loader = [1, 2]
    engine = Engine(update_fn)

    pbar = ProgressBar(ncols=80)
    pbar.attach(engine, output_transform=lambda x: x)

    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    if get_tqdm_version() < Version("4.49.0"):
        expected = "Epoch [2/2]: [1/2]  50%|█████     , output=1 [00:00<00:00]"
    else:
        expected = "Epoch [2/2]: [1/2]  50%|██████████████████                  , output=1 [00:00<?]"
    assert err[-1] == expected


def test_pbar_with_str_output(capsys):
    n_epochs = 2
    loader = [1, 2]
    engine = Engine(update_fn)

    pbar = ProgressBar(ncols=80)
    pbar.attach(engine, output_transform=lambda x: "red")

    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    if get_tqdm_version() < Version("4.49.0"):
        expected = "Epoch [2/2]: [1/2]  50%|█████     , output=red [00:00<00:00]"
    else:
        expected = "Epoch [2/2]: [1/2]  50%|█████████████████                 , output=red [00:00<?]"
    assert err[-1] == expected


def test_pbar_with_tqdm_kwargs(capsys):
    n_epochs = 10
    loader = [1, 2, 3, 4, 5]
    engine = Engine(update_fn)

    pbar = ProgressBar(desc="My description: ", ncols=80)
    pbar.attach(engine, output_transform=lambda x: x)
    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    expected = "My description:  [10/10]: [4/5]  80%|███████████████▏   , output=1 [00:00<00:00]"
    assert err[-1] == expected


def test_pbar_for_validation(capsys):
    loader = [1, 2, 3, 4, 5]
    engine = Engine(update_fn)

    pbar = ProgressBar(desc="Validation", ncols=80)
    pbar.attach(engine)
    engine.run(loader, max_epochs=1)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    expected = "Validation: [4/5]  80%|██████████████████████████████████▍         [00:00<00:00]"
    assert err[-1] == expected


def test_pbar_output_tensor(capsys):
    def _test(out_tensor, out_msg):
        loader = [1, 2, 3, 4, 5]

        def update_fn(engine, batch):
            return out_tensor

        engine = Engine(update_fn)

        pbar = ProgressBar(desc="Output tensor", ncols=80)
        pbar.attach(engine, output_transform=lambda x: x)
        engine.run(loader, max_epochs=1)

        captured = capsys.readouterr()
        err = captured.err.split("\r")
        err = list(map(lambda x: x.strip(), err))
        err = list(filter(None, err))
        expected = f"Output tensor: [4/5]  {out_msg} [00:00<00:00]"
        assert err[-1] == expected

    _test(out_tensor=torch.tensor([5, 0]), out_msg="80%|████████████▊   , output_0=5, output_1=0")
    _test(out_tensor=torch.tensor(123), out_msg="80%|██████████████████████▍     , output=123")
    _test(out_tensor=torch.tensor(1.234), out_msg="80%|█████████████████████▌     , output=1.23")


def test_pbar_output_warning(capsys):
    loader = [1, 2, 3, 4, 5]

    def update_fn(engine, batch):
        return torch.zeros(1, 2, 3, 4)

    engine = Engine(update_fn)

    pbar = ProgressBar(desc="Output tensor", ncols=80)
    pbar.attach(engine, output_transform=lambda x: x)
    with pytest.warns(UserWarning):
        engine.run(loader, max_epochs=1)


def test_pbar_on_epochs(capsys):
    n_epochs = 10
    loader = [1, 2, 3, 4, 5]
    engine = Engine(update_fn)

    pbar = ProgressBar(ncols=80)
    pbar.attach(engine, event_name=Events.EPOCH_STARTED, closing_event_name=Events.COMPLETED)
    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    actual = err[-1]
    expected = "Epoch: [9/10]  90%|██████████████████████████████████████████▎     [00:00<00:00]"
    assert actual == expected


def test_pbar_with_max_epochs_set_to_one(capsys):
    n_epochs = 1
    loader = [1, 2]
    engine = Engine(update_fn)

    pbar = ProgressBar(ncols=80)
    pbar.attach(engine, ["a"])

    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    if get_tqdm_version() < Version("4.49.0"):
        expected = "Iteration: [1/2]  50%|█████     , a=1 [00:00<00:00]"
    else:
        expected = "Iteration: [1/2]  50%|█████████████████████▌                     , a=1 [00:00<?]"
    assert err[-1] == expected


def test_pbar_wrong_events_order():
    engine = Engine(update_fn)
    pbar = ProgressBar(ncols=80)

    with pytest.raises(ValueError, match="should be called before closing event"):
        pbar.attach(engine, event_name=Events.COMPLETED, closing_event_name=Events.COMPLETED)

    with pytest.raises(ValueError, match="should be called before closing event"):
        pbar.attach(engine, event_name=Events.COMPLETED, closing_event_name=Events.EPOCH_COMPLETED)

    with pytest.raises(ValueError, match="should be called before closing event"):
        pbar.attach(engine, event_name=Events.COMPLETED, closing_event_name=Events.ITERATION_COMPLETED)

    with pytest.raises(ValueError, match="should be called before closing event"):
        pbar.attach(engine, event_name=Events.EPOCH_COMPLETED, closing_event_name=Events.EPOCH_COMPLETED)

    with pytest.raises(ValueError, match="should be called before closing event"):
        pbar.attach(engine, event_name=Events.ITERATION_COMPLETED, closing_event_name=Events.ITERATION_STARTED)

    with pytest.raises(ValueError, match="should not be a filtered event"):
        pbar.attach(engine, event_name=Events.ITERATION_STARTED, closing_event_name=Events.EPOCH_COMPLETED(every=10))


def test_pbar_with_nan_input():
    def update(engine, batch):
        x = batch
        return x.item()

    def create_engine():
        engine = Engine(update)
        pbar = ProgressBar(ncols=80)

        engine.add_event_handler(Events.ITERATION_COMPLETED, TerminateOnNan())
        pbar.attach(engine, event_name=Events.EPOCH_COMPLETED, closing_event_name=Events.COMPLETED)
        return engine

    data = torch.from_numpy(np.array([np.nan] * 25))
    engine = create_engine()
    engine.run(data)
    assert engine.should_terminate
    assert engine.state.iteration == 1
    assert engine.state.epoch == 1

    data = torch.from_numpy(np.array([1] * 1000 + [np.nan] * 25))
    engine = create_engine()
    engine.run(data)
    assert engine.should_terminate
    assert engine.state.iteration == 1001
    assert engine.state.epoch == 1


def test_pbar_on_callable_events(capsys):
    n_epochs = 1
    loader = list(range(100))
    engine = Engine(update_fn)

    pbar = ProgressBar(ncols=80)
    pbar.attach(engine, event_name=Events.ITERATION_STARTED(every=10), closing_event_name=Events.EPOCH_COMPLETED)
    engine.run(loader, max_epochs=n_epochs)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    actual = err[-1]
    expected = "Iteration: [90/100]  90%|████████████████████████████████████▉     [00:00<00:00]"
    assert actual == expected


def test_tqdm_logger_epoch_length(capsys):
    loader = list(range(100))
    engine = Engine(update_fn)
    pbar = ProgressBar(persist=True, ncols=80)
    pbar.attach(engine)
    engine.run(loader, epoch_length=50)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    actual = err[-1]
    expected = "Iteration: [50/50] 100%|██████████████████████████████████████████ [00:00<00:00]"
    assert actual == expected


def test_tqdm_logger_iter_without_epoch_length(capsys):
    size = 11

    def finite_size_data_iter(size):
        for i in range(size):
            yield i

    def train_step(trainer, batch):
        pass

    trainer = Engine(train_step)

    @trainer.on(Events.ITERATION_COMPLETED(every=size))
    def restart_iter():
        trainer.state.dataloader = finite_size_data_iter(size)

    pbar = ProgressBar(persist=True, ncols=80)
    pbar.attach(trainer)

    data_iter = finite_size_data_iter(size)
    trainer.run(data_iter, max_epochs=5)

    captured = capsys.readouterr()
    err = captured.err.split("\r")
    err = list(map(lambda x: x.strip(), err))
    err = list(filter(None, err))
    actual = err[-1]
    expected = "Epoch [5/5]: [11/11] 100%|████████████████████████████████████████ [00:00<00:00]"
    assert actual == expected
